/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */

#pragma runopts(POSIX(ON))

#define DTL_LOG CoZAgent_LOG

#define _OPEN_THREADS
//The following includes must be in this order for CoZ
#include "CoZConstants.h"
#include "DtlLog.h"
#include "sys_util.h"

#include "pthread_util.h"
#include "CoZUtil.h"
#include <sys/stat.h>
#ifdef HAVE_IO_H
#include <io.h>
#endif

#include "CoZAgent.h"

#ifndef HOST_NAME_MAX
#define HOST_NAME_MAX 255
#endif

DtlLog CoZAgent_LOG("CoZAgent");

static const int CHILD_WAIT_SECONDS_MIN = 2;
static const int CHILD_WAIT_SECONDS_DEFAULT = 20;

const char* CMD_VERSION	= "/VERSION";
const char* CMD_SETENV	= "/SETENV";
const char* CMD_START	= "/START";
const char* CMD_QUIESCE	= "/QUIESCE";
const char* CMD_KILL	= "/KILL";
const char* CMD_CMD		= "/CMD";

const char* DEFAULT_SHELL = "/bin/sh";

#define Str(x) #x
#define Xstr(x) Str(x)
const char* COZ_VERSION = Xstr(BUILD_VERSION) ;

extern int errno;

void usage() {
	fprintf(stderr,
"usage: cozagent [OPTION]...\n"
"\n"
"options:\n"
"  -c -> enable MVS console read after stdin\n"
"  -q -> do not display periodic 'Waiting for completion' messages\n"
"  -h -> display this help and exit\n"
"  -L M|A|C|E|W|N|I|D|T  -> sets the default logging to the given level: \n"
"     (eMerg, Alert, Crit, Err, Warning, Info, Notice, Debug, Trace) \n"
"  -L logname=level -> changes the level of a specific logname(component) \n"
"  -L e  -> prefixes log messages with the elapsed time in milliseconds \n"
"  -L s  -> messages are logged to SYSLOG facility instead of stderr\n"
"  -L f=filename  -> writes log messages a given file instead of stderr \n"
"  -n -> run the target-command without a shell \n"
"  -v -> display the current version and exit\n"
"  -w secs -> seconds to wait for child process to terminate (default=20)\n"
	);
}

int main(int argc, char** argv) {

	DtlLog::setProgName(basename(argv[0]));
	DtlLog* defaultLog = DtlLog::getDefaultLog();
	CoZAgent_LOG.printLogName(false); // since redundant with prog name
	defaultLog->logmsg(LOG_NOTICE, "version: %s", COZ_VERSION);

	CoZAgent cozAgent(argv[0]);

	int opt;
	extern char* optarg;
	extern int optopt;
	int wait_seconds = 0;
	while ((opt = getopt(argc, argv, ":cDhL:nqvw:")) != -1) {
		switch (opt) {
			case 'c':
				cozAgent.setReadConsole(true);
				break;
    		case 'D':
   				cozAgent.setDebugEnabled(true);
    			break;
    		case 'h':
    			usage();
    			exit(RC_OK);
    		case 'L':
   				DtlLog::handleLogOption(optarg);
    			break;
    		case 'n':
    			cozAgent.setNoShell(true);
    			break;
    		case 'q':
    			cozAgent.setQuietMode(true);
    			break;
    		case 'v':
    			printf("%s\n", COZ_VERSION);
    			exit(RC_OK);
    		case 'w':
    			wait_seconds = atoi(optarg);
    			if (wait_seconds < CHILD_WAIT_SECONDS_MIN) {
    				LOG(LOG_ERR, "Argument to option %c must be >= %d", optopt, CHILD_WAIT_SECONDS_MIN);
    				exit(RC_INPUT_ERROR);    				
    			}
    			cozAgent.setChildWaitSeconds(wait_seconds);
    			break;
    		case ':':
				LOG(LOG_ERR, "Missing argument for option: %c", optopt);
				exit(RC_INPUT_ERROR);
    			break;
			default:
				LOG(LOG_ERR, "Unknown option: %c", optopt);
				usage();
				exit(RC_INPUT_ERROR);
    	}
    }
	LOG(LOG_INFO, "version: %s", COZ_VERSION);
	LOG(LOG_INFO, "Copyright (C) Dovetailed Technologies, LLC. 2006-2017. All rights reserved.");

    return cozAgent.run();
}

CoZAgent::CoZAgent(char* cmd_path) :
	debug_active(false),
	debug_enabled(false),
	read_console(false),
	quiet_mode(false),
	no_shell(false),
	child_wait_seconds(CHILD_WAIT_SECONDS_DEFAULT),
	target_command(NULL),
	fullpath(cmd_path),
	progname(basename(cmd_path)),
	target_pid(0),
	stdin_dd_pid(0),
	stdout_dd_pid(0),
	stderr_dd_pid(0),
	console_tid(pthread_self()),
	orphan_detection_tid(pthread_self())
{
	TRACE_CTX("CoZAgent");

	stdin_dd_pipe = new int[2];
	Pipe(stdin_dd_pipe);

	stdout_dd_pipe = new int[2];
	Pipe(stdout_dd_pipe);

	stderr_dd_pipe = new int[2];
	Pipe(stderr_dd_pipe);

}

CoZAgent::~CoZAgent() {
	delete[] target_command;
	delete[] stdin_dd_pipe;
	delete[] stdout_dd_pipe;
	delete[] stderr_dd_pipe;
}

void CoZAgent::configurePath() {
	TRACE_CTX("configurePath");
	char* path = getenv("PATH");
	char* fullpath2 = fullpath ? strdup(fullpath) : NULL;  // dirname() may modify/return
	char* dir = dirname(fullpath2);
	char* new_path = new char[strlen(path) + strlen(dir) + 2];
	LOG(LOG_DEBUG, "cozagent dir: %s", dir);
	char cwd[256];
	if (!getcwd(cwd, sizeof cwd)) {
		LOG(LOG_ERR,"getcwd() failed: %s", strerror(errno));
		throw RC_AGENT_ERROR;
	}
	LOG(LOG_DEBUG, "cwd: %s", cwd);
	strcpy(new_path, path);
	strcat(new_path, ":");
	strcat(new_path, dir);
	Setenv("PATH", new_path, 1);
	LOG(LOG_DEBUG, "PATH:");
	char* d = strtok(new_path, ":;");
	while (d) {
		LOG(LOG_DEBUG, "\t%s", d);
		d = strtok(NULL, ":;");
	}
	if (fullpath2) free(fullpath2);
	delete[] new_path;
}

bool CoZAgent::matchCommand(const char* command, char* str) {
	return !strncasecmp(command, str, strlen(command));
}

void CoZAgent::initialize() {
	TRACE_CTX("initialize");
	char buf[1024];
	configurePath();

	// Initialize COZ_SERVER_PROTOCOL_VERSION to 1.  Later versions of CoZLauncher will
	// Send an environment variable with the protocol version
	Setenv(ENV_COZ_SERVER_PROTOCOL_VERSION, "1", 1);

	bool start_received = false;
	LOG(LOG_DEBUG, "Waiting for configuration commands...");
	while (Fgets(buf, sizeof(buf), stdin)) {
		int buf_len = strlen(buf);
		if (buf[buf_len-1] == '\n') {
			buf[--buf_len] = '\0';
		}
		if (matchCommand(CMD_SETENV, buf)) {
			setEnvironmentVariable(CoZUtil::trim(buf+strlen(CMD_SETENV)+1));
		} else if (matchCommand(CMD_VERSION, buf)) {
			printf("%s%s\n", COZ_AGENT_VERSION_MSG, COZ_VERSION);
		} else if (matchCommand(CMD_START, buf)) {
			start_received = true;
			char* p = CoZUtil::trim(buf+strlen(CMD_START)+1);
			if (strlen(p) == 0) {
				target_command = NULL;
			} else {
				target_command = new char[strlen(p) + 1];
				strcpy(target_command, p);
			}
			LOG(LOG_DEBUG, "target_command='%s'", target_command ? target_command : "NULL");
			break;
		} else {
			LOG(LOG_ERR, "Unexpected agent init command: %s", buf);
			throw RC_AGENT_ERROR;
		}
	}
	if (!start_received) {
			LOG(LOG_ERR, "No %s command received", CMD_START);
			throw RC_AGENT_ERROR;
	}
}

bool CoZAgent::isChildActive(pid_t pid) {
	if (!pid) return false;
	int stat;
	return !Waitpid(pid, &stat, WNOHANG | WUNTRACED);
}

void CoZAgent::killChild(const char* child_name, pid_t& pid, int signal) {
	if (pid <= 0) {
		LOG(LOG_DEBUG,"Child process %s not killed - pid was: %d", child_name, pid);
		return; // we didn't really have a pid
	}
	LOG(LOG_DEBUG,"Killing child process %s (%d) with signal %d", child_name, pid, signal);
	kill(pid, signal);
	LOG(LOG_DEBUG,"Waiting for child %s (%d) to terminate", child_name, pid);
	int stat;
	pid = Waitpid(pid, &stat, 0);
	if (pid != -1) waitpidCleanup(child_name, pid, stat);
}

int CoZAgent::run() {
	struct sigaction sig_alarm_action; //Used by keep alive timer
	TRACE_CTX("run");
	int rc = 0;
	if (debug_enabled) {
		LOG(LOG_NOTICE,"Waiting for debugger attachment - PID=%d.  When attached, set debug_active=true.",getpid());
		while (!debug_active) {
			sleep(5);
		}
	}
	try {
		initialize();
		startTargetProgram();
		startStdinDDReader();
		startStdoutDDWriter();
		startStderrDDWriter();
		if (read_console) {
			usePipe(stdin_dd_pipe, false, true);
			Pthread_create(&console_tid, NULL, consoleReaderThread, this);
		} else {
			usePipe(stdin_dd_pipe, false, false);
		}
		usePipe(stdout_dd_pipe, false, false);
		usePipe(stderr_dd_pipe, false, false);

		if (!quiet_mode) {
			// we set up an interval alarm used to keep the connection alive
			memset(&sig_alarm_action, 0, sizeof(sig_alarm_action));
			sig_alarm_action.sa_handler = sig_alarm_handler;
			sig_alarm_action.sa_flags = SA_RESTART;
			if (sigaction(SIGALRM, &sig_alarm_action, NULL) == -1) {
				LOG(LOG_WARNING, "Keep alive timer could not be established.  No keep alive messages will be sent.");
			} else {
				alarm(KEEP_ALIVE_DELAY_SECONDS);
			}
		}

		Pthread_create(&orphan_detection_tid, NULL, orphanDetectionThread, NULL);

		rc = waitForChildren();

	} catch (const int thrown_rc) {
		rc = thrown_rc;
	} catch (...) {
		LOG(LOG_ERR, "caught unexpected exception");
    	rc = RC_SYSTEM_ERROR;
	}

	signal(SIGALRM, SIG_IGN);  // ignore any pending alarms
	//Always write the completion message, regardless of logging level
	//The launcher expects this message in all cases.
	fprintf(stdout,"%s%d\n",COZ_AGENT_COMPLETION_MSG, rc);
	fflush(stdout);
	TRACE_CTX_RETVAL("%d", rc);
	return rc;
}

void CoZAgent::setEnvironmentVariable(char* var) {
	TRACE_CTX("setEnvironmentVariable");
	char* name;
	char* value;
	name = strtok(var, "=");
	value = strtok(NULL, "");
	Setenv(name, value, 1);
	LOG(LOG_DEBUG,"Set environment variable: '%s'",name);
}

/*
 * Use fromdsn to read the target program's stdin from DD://STDIN.
 * In this case, fromdsn's:
 * 		stdin comes from DD://STDIN
 * 		stderr is tied to our stderr
 * 		stdout is piped to the target program on the stdin_dd_pipe.
 */
void CoZAgent::startStdinDDReader() {
	TRACE_CTX("startStdinDDReader");

    pid_t pid = Fork();
	if (pid) {
		if (pid > 0) {
			stdin_dd_pid = pid;
	   		LOG(LOG_DEBUG, "stdin_dd_pid = %d", stdin_dd_pid);
		}
		return;
	}

	usePipe(stdin_dd_pipe, false, true);
	usePipe(stdout_dd_pipe, false, false);
	usePipe(stderr_dd_pipe, false, false);
	Dup2(stdin_dd_pipe[WRITE_END], STDOUT_FD);

	char* cmd[] = {
			const_cast<char*>("fromdsn"), 
			const_cast<char*>("-sock"), 
			const_cast<char*>("//DD:STDIN"), 
			NULL};
	execvp(cmd[0], cmd);
	LOG(LOG_ERR, "execvp: %s", strerror(errno));
	exit(RC_SYSTEM_ERROR);

}

/*
 * Use todsn to write the target program's stdout to DD://STDOUT.
 * In this case, todsn's:
 * 		stdin is piped to us from the target program stdout on stdout_dd_pipe
 * 		stderr is tied to our stderr
 * 		stdout is sent to DD://STDOUT
 */
void CoZAgent::startStdoutDDWriter() {
	TRACE_CTX("startStdoutDDWriter");

    pid_t pid = Fork();
	if (pid) {
		if (pid > 0) {
			stdout_dd_pid = pid;
	   		LOG(LOG_DEBUG, "stdout_dd_pid = %d", stdout_dd_pid);
		}
		return;
	}
	//todsn does not use STDOUT, so we close here to avoid interference with
	//the parent's STDIN processing. (specifically in the terminal I/O case,
	//stdin, stdout and sterr share a device)
	Close(STDOUT_FD);

	usePipe(stdin_dd_pipe, false, false);
	usePipe(stdout_dd_pipe, true, false);
	usePipe(stderr_dd_pipe, false, false);
	Dup2(stdout_dd_pipe[READ_END], STDIN_FD);

	char* cmd[] = {
			const_cast<char*>("todsn"), 
			const_cast<char*>("-sock"), 
			const_cast<char*>("-z"), 
			const_cast<char*>("//DD:STDOUT"), 
			NULL};
	execvp(cmd[0], cmd);
	/* should not get here */
	LOG(LOG_ERR, "execvp: %s", strerror(errno));
	exit(RC_SYSTEM_ERROR);
}

/*
 * Use todsn to write the target program's stderr to DD://STDOUT.
 * In this case, todsn's:
 * 		stdin is piped to us from the target program stderr on stderr_dd_pipe
 * 		stderr is tied to our stderr
 * 		stdout is sent to DD://STDERR
 */
void CoZAgent::startStderrDDWriter() {
	TRACE_CTX("startStderrDDWriter");

    pid_t pid = Fork();
	if (pid) {
		if (pid > 0) {
			stderr_dd_pid = pid;
	   		LOG(LOG_DEBUG, "stderr_dd_pid = %d", stderr_dd_pid);
		}
		return;
	}
	//todsn does not use STDOUT, so we close here to avoid interference with
	//the parent's STDIN processing. (specifically in the terminal I/O case,
	//stdin, stdout and sterr share a device)
	Close(STDOUT_FD);

	usePipe(stdin_dd_pipe, false, false);
	usePipe(stdout_dd_pipe, false, false);
	usePipe(stderr_dd_pipe, true, false);
	Dup2(stderr_dd_pipe[READ_END], STDIN_FD);

	char* cmd[] = {
			const_cast<char*>("todsn"), 
			const_cast<char*>("-sock"), 
			const_cast<char*>("-z"), 
			const_cast<char*>("//DD:STDERR"), 
			NULL};
	execvp(cmd[0], cmd);
	/* should not get here */
	LOG(LOG_ERR, "execvp: %s", strerror(errno));
	exit(RC_SYSTEM_ERROR);
}


void CoZAgent::startTargetProgram() {
	TRACE_CTX("startTargetProgram");

	const char* shell_path = DEFAULT_SHELL;
	struct passwd *pw;
	pw = getpwuid(geteuid());
	LOG(LOG_DEBUG, "getpwuid(geteuid()) pw=%p, pw_shell=%s", pw, (pw && pw->pw_shell) ? pw->pw_shell : "NULL");
	if (pw && pw->pw_shell && strlen(pw->pw_shell) > 0) {
		shell_path = pw->pw_shell;
	}

	pid_t pid = Fork();
	if (pid) {
		if (pid > 0) {
			target_pid = pid;
			char host_name[HOST_NAME_MAX];
			gethostname(host_name, sizeof(host_name));
	   		LOG(LOG_DEBUG, "target_pid = %d", target_pid);
	   		fprintf(stdout,"CoZAgent: %s@%s target program '%s' PID: %d\n",
	   				(pw) ? pw->pw_name : "<unknown>",
	   				host_name,
	   				target_command ? target_command : shell_path,
	   				pid);
		}
		return;
	}

	usePipe(stdin_dd_pipe, true, false);
	usePipe(stdout_dd_pipe, false, true);
	usePipe(stderr_dd_pipe, false, true);
	Dup2(stdin_dd_pipe[READ_END], STDIN_FD);
	Dup2(stdout_dd_pipe[WRITE_END], STDOUT_FD);
	Dup2(stderr_dd_pipe[WRITE_END], STDERR_FD);

    #ifdef TARGET_CYGWIN
		setmode(STDIN_FD, O_TEXT);
		setmode(STDOUT_FD, O_TEXT);
		setmode(STDERR_FD, O_TEXT);
	#endif

	char* argv[256];
	const int num_argv = 256;

	//The shell name to be passed in argv[0] is preceded by '-' to request
	//a login shell
	argv[0] = new char[strlen(shell_path) + 2];
	strcpy(argv[0],"-");
	strcat(argv[0],shell_path);

	if (!target_command) {
		//No command - run the default shell
		argv[1] = NULL;
		LOG(LOG_DEBUG,"About to execute shell %s", argv[0]);
		execv(shell_path, argv);
		LOG(LOG_ERR, "execv: %s - %s", argv[0], strerror(errno));
		exit(RC_SYSTEM_ERROR);

	} else if (no_shell) {
		// run the target_command naked, but first need to tokenize
		char* cmd = strdup(target_command);
		argv[0] = strtok(cmd, " ");
		int i;
		for (i=1; i<num_argv; i++) {
			argv[i] = strtok(NULL, " ");
			if (argv[i] == NULL)
				break;
		}
		if (i == num_argv) {  // too many tokens
			LOG(LOG_ERR, "Too many tokens in target-command: %s", target_command);
			exit(RC_INPUT_ERROR);
		}
		LOG(LOG_DEBUG,"About to execute target-command: %s", target_command);
		execv(argv[0], argv);
		LOG(LOG_ERR, "execv: %s - %s", argv[0], strerror(errno));
		exit(RC_SYSTEM_ERROR);

	} else {
		// execute <shell> -c <target-program>
		argv[1] = const_cast<char*>("-c");
		argv[2] = target_command;
		argv[3] = NULL;
		LOG(LOG_DEBUG,"About to execute %s %s %s", argv[0], argv[1], argv[2]);
		execv(shell_path, argv);
		LOG(LOG_ERR, "execv: %s %s %s - (%d) %s", argv[0], argv[1], argv[2], errno, strerror(errno));
		exit(RC_SYSTEM_ERROR);
	}
}

void CoZAgent::usePipe(int* pipe, bool useReadEnd, bool useWriteEnd) {
	if (!useReadEnd) Close(pipe[READ_END]);
	if (!useWriteEnd) Close(pipe[WRITE_END]);
}

int CoZAgent::waitForChildren() {
	TRACE_CTX("waitForChildren");
    int stat;
    int rc = RC_TARGET_KILLED;
    bool stdio_failure = false;

    //Wait for target program, then make sure DD pids are finished
	pid_t pid = Waitpid(target_pid, &stat, 0);
   	rc = waitpidCleanup("Target Program", target_pid, stat);

    //Wait for all DD readers/writers to drain.
    int retry_count = 10;
    int elapsed_wait_seconds = 2;
    while (stdin_dd_pid || stdout_dd_pid || stderr_dd_pid) {
    	pid_t pid = Waitpid(-1, &stat, WNOHANG | WUNTRACED);
   		if (pid == 0) {
			//At least one of the children is still running.  For a total of two
   			//seconds, use a 200000 microsecond sleep interval.  After two seconds
   			//and up to child_wait_seconds, use a 1 second sleep interval.
   			if (retry_count) {
   				retry_count--;
   				usleep(200000);
   			} else if (elapsed_wait_seconds < child_wait_seconds) {
   				elapsed_wait_seconds++;
   				sleep(1);
   			} else {
   	   			break;
   			}
   			LOG(LOG_DEBUG,"At least one child I/O process still running, waiting again");
   		} else if (pid == stdin_dd_pid && waitpidCleanup("STDIN DD Reader", stdin_dd_pid, stat, true)) {
   			stdio_failure = true;
   		} else if (pid == stdout_dd_pid && waitpidCleanup("STDOUT DD Writer", stdout_dd_pid, stat, true)) {
			stdio_failure = true;
   		} else if (pid == stderr_dd_pid && waitpidCleanup("STDERR DD Writer", stderr_dd_pid, stat, true)) {
   			stdio_failure = true;
   		}
	}
    
	//Kill all children that haven't reported
	if (stdin_dd_pid) {
		LOG(LOG_WARNING,"STDIN DD Reader did not complete within the timeout period of %d seconds; killing", child_wait_seconds);
		killChild("STDIN DD Reader", stdin_dd_pid, SIGTERM);
		stdio_failure = true;
	}
	if (stdout_dd_pid) {
		LOG(LOG_WARNING,"STDOUT DD Writer did not complete within the timeout period of %d seconds; killing", child_wait_seconds);
		killChild("STDOUT DD Writer", stdout_dd_pid, SIGTERM);
		stdio_failure = true;
	}
	if (stderr_dd_pid) {
		LOG(LOG_WARNING,"STDERR DD Writer did not complete within the timeout period of %d seconds; killing", child_wait_seconds);
		killChild("STDERR DD Writer", stderr_dd_pid, SIGTERM);
		stdio_failure = true;
	}

	//If the target program does not report an error, but one of the
	//std stream redirectors fails or had to be killed, set a failure code for the agent.
	if (!rc && stdio_failure) {
		rc = RC_IO_ERROR;
	}
	
	if (!pthread_equal(console_tid, pthread_self())) {
		Pthread_cancel(console_tid);
	}

	TRACE_CTX_RETVAL("rc=%d", rc);
    return rc;
}

int CoZAgent::waitpidCleanup(const char* child, pid_t& pid, int stat, bool ignoreSigPipe) {
	TRACE_CTX("waitpidCleanup");
    int rc = RC_OK;
    int lvl;
    if (WIFEXITED (stat)) {
    	rc = WEXITSTATUS (stat);
		LOG((rc) ? LOG_ERR : LOG_DEBUG, "%s(%d) ended with RC=%d", child, pid, rc);
    } else if (WIFSIGNALED (stat)) {
    	if (WTERMSIG (stat) == SIGPIPE && ignoreSigPipe) {
    		LOG(LOG_INFO, "%s(%d) received SIGPIPE, which will be ignored", child, pid);
    	} else {
    		LOG(LOG_ERR, "%s(%d) received signal - %d", child, pid, WTERMSIG (stat));
    		rc = RC_TARGET_KILLED;
    	}
   	} else if (WIFSTOPPED (stat)) {
   		LOG(LOG_ERR, "%s(%d) stopped on signal - %d", child, pid, WSTOPSIG (stat));
	} else {
		LOG(LOG_ERR, "Unknown status code for %s(%d): %i", child, pid, stat);
    }
    pid = 0; //reset so we don't wait for it again
    return rc;
}

/*
 * Read and process commands sent from the MVS console.
 * Commands directed at the agent itself are sent with a preceeding slash (/)
 * Commands and data directed at the target program are unadorned.
 */
void* consoleReaderThread(void* pCoZAgent) {
	TRACE_CTX("consoleReaderThread");
	ssize_t nread = 0;
	int stat;
	CoZAgent* cozAgent = (CoZAgent*) pCoZAgent;
	char buf[1024];
	while (Fgets(buf, sizeof(buf), stdin)) {
		int buf_len = strlen(buf);
		if (buf[buf_len-1] == '\n') {
			buf[--buf_len] = '\0';
		}
		char* cmd = CoZUtil::trim(buf);
		LOG(LOG_DEBUG, "Received from console: '%s'", cmd);
		if (!cmd || strlen(cmd) == 0) {
			LOG(LOG_DEBUG, "NULL command received from console.");
			continue;
		}
		if (cozAgent->matchCommand(CMD_QUIESCE, cmd)) {
			LOG(LOG_DEBUG, "%s received - closing target program STDIN", CMD_QUIESCE);
			Close(cozAgent->stdin_dd_pipe[WRITE_END]);
		} else if (cozAgent->matchCommand(CMD_KILL, cmd)) {
			char* sig_level = CoZUtil::trim(buf+strlen(CMD_KILL));
			int sig = (strlen(sig_level)) ? atoi(sig_level) : SIGKILL;
			LOG(LOG_DEBUG, "%s received - signaling target program with signal=%d", CMD_KILL, sig);
			cozAgent->killChild("Target Program", cozAgent->target_pid, sig);
		} else if (cozAgent->matchCommand(CMD_CMD, cmd)) {
			char* agent_cmd = CoZUtil::trim(cmd+strlen(CMD_CMD));
			LOG(LOG_DEBUG, "%s %s received - issuing system call", CMD_CMD, agent_cmd);
			int stat = system(agent_cmd);
			if (stat == -1) {
				LOG(LOG_ERR,"system(%s) failed - [%d]%s", agent_cmd, errno, strerror(errno));
			}
			int rc = WEXITSTATUS(stat);
			if (rc) { //If non-zero RC, write back to the console via stdout
				fprintf(stdout, "%s exited with RC=%d", cozAgent->progname, rc);
			}
		} else if (cmd[0] == '/') {
			fprintf(stdout, "%s: Unrecognized command '%s' received", cozAgent->progname, cmd);
		} else {
			if (!cozAgent->isStdinDrained()) {
				fprintf(stdout,"%s: Target program is still reading STDIN.  Command %s ignored", cozAgent->progname, cmd);
			} else {
				Write(cozAgent->stdin_dd_pipe[WRITE_END], cmd, strlen(cmd));
			}
		}
	}
	Close(cozAgent->stdin_dd_pipe[WRITE_END]);
	pthread_exit(NULL);
	//Unreachable, but required
	return NULL;
}

/*
 * Periodically check to make sure that we have a parent (sshd).  If not, assume
 * the sshd connection has been lost and kill the process group to avoid leaving around
 * orphaned processes.
 */
void* orphanDetectionThread(void* pCoZAgent) {
	TRACE_CTX("orphanDetectionThread");
	while (true) {
		sleep(1);
		if (getppid() == 1) {
			killpg(getpgrp(), SIGTERM);
			break;
		}
	}
	pthread_exit(NULL);
	//Unreachable, but required
	return NULL;
}

/**
 * SIGALRM handler to keep the connection alive by writing something to stderr
 * No alarms are configured if quiet_mode.
 */
void sig_alarm_handler(int sig_type) {
	time_t t = time(NULL);
	struct tm local_time;
	localtime_r(&t, &local_time);
	fprintf(stderr, "%s [%2.2d:%2.2d] Waiting for completion...\n",
			DtlLog::getDefaultLog()->getProgName(),
			local_time.tm_hour,
			local_time.tm_min);
	signal(SIGALRM, sig_alarm_handler); // Re-establish handler
	alarm(CoZAgent::KEEP_ALIVE_DELAY_SECONDS);  // start an alarm to replace this one
}
