/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */

// Enable signal handling support, etc...
#pragma runopts(POSIX(ON))

// must be before any includes of "DtlLog.h"
#define DTL_LOG CoZClient_LOG

//The following includes must be in this order for CoZ
#include "CoZConstants.h"
#include "DtlLog.h"
#include "sys_util.h"

#include "CoZPacket.h"
#include "CoZClient.h"
#include "CoZUtil.h"
#include <sys/stat.h>  /* for fstat() */
#include <sys/un.h>
#include <dirent.h>
#ifdef TARGET_CYGWIN
#include <io.h>
#endif

DtlLog CoZClient_LOG("CoZClient");  // definition of our log

volatile bool termination_requested = false;

struct sigaction sVal;
pid_t 	ssh_pid = 0;   						// pid of child ssh command needed by a signal handler

static CoZBaseClient* pCoZClient = NULL;

/**
 Changed in 1.2.1:
 Improve SIGPIPE error handling in CoZAgent and CoZClient -
  - CoZClient (and fromdsn/todsn) will explicitly handle EPIPE
    errors on sockets and standard handles.
  - CoZClient will exit=103 for EPIPE errors on sockets,
    but write error message and signal SIGPIPE for errors
    on stdin/stdout/stderr
  - CoZAgent will log a info message for SIGPIPE completions
    on stdin/stdout/stderr DD writers, but will not exit with
    and error in these cases.
 */

/**
 * signal handler to force socket shutdown.
 */
void sig_term_handler(int sig_type) {
	TRACE_CTX("sig_term_handler");
	if (pCoZClient) {
		pCoZClient->shutdownSocket();
	}
}

void kill_ssh_child() {
	int stat;
	if (ssh_pid <= 0)
		return;
	LOG(LOG_DEBUG,"Killing SSH child process (%d) with signal %d", ssh_pid, SIGTERM);
	if (kill(ssh_pid, SIGTERM) == -1) {
		LOG(LOG_DEBUG,"kill() failed for SSH child process (%d): [%d] %s", ssh_pid, errno, strerror(errno));
	}
	waitpid(ssh_pid, &stat, 0);  // Wait, but don't report any errors because we are shutting down.
}

/**
 * signal handler to ssh connection termination.
 */
void sig_child_handler(int sig_type, siginfo_t *si, void *context) {
	TRACE_CTX("sig_child_handler");
	int stat;

	if (si->si_pid != ssh_pid) {
	   LOG(LOG_TRACE, "leaving sig_child_handler(ignoring) ");
	   return;
	}

	if (!termination_requested) {
		LOG(LOG_TRACE, "CoZServer terminated connection");
		termination_requested = true;
		if (ssh_pid > 0) waitpid(ssh_pid, &stat, 0);  // Wait, but don't report any errors because we are shutting down.
	}
}

/**
 * Constructor/ initializers
 */
CoZBaseClient::CoZBaseClient() :
		cmd_argc(0),
		cmd_argv(NULL),
		return_code(0),
		server_exitcode_received(false),
	    server_exitstatus_received(false),
		conn_type(CONN_TYPE_SOCKET),
		io_config_stdin(IO_CONFIG_TEXT),
		io_config_stdout(IO_CONFIG_TEXT),
		io_config_stderr(IO_CONFIG_TEXT),
		io_config_errpt(IO_CONFIG_ERRP_INCLUDE),
		ssh_cmd(NULL),
		user_host(NULL),
		tunneled(false),
		sock(0),
		socket_server_host(NULL),
		socket_server_port(NULL),
		socket_auth_token(NULL),
		fd_conn_snd(0),
		fd_conn_rcv(0),
		send_buf(NULL),
		rcv_buf(NULL),
		server_retcode(0),
		server_errno(0),
    	server_errno2(0),
    	server_msg(NULL),
    	server_errno_msg(NULL),
    	server_err_component(NULL),
    	total_bytes_received(0),
    	total_bytes_sent(0),
    	server_bytes_received(0),
    	server_bytes_sent(0),
    	clientCodePage(NULL),
		serverProtocolVersion(-1),
		protocolVersion(COZ_PROTOCOL_VERSION),
		socket_file_name(NULL)
{
	negotiateProtocolVersion();
}


/**
 * Destructor
 */
CoZBaseClient::~CoZBaseClient() {

	if (cmd_argv) delete[] cmd_argv;
	if (ssh_cmd) delete[] ssh_cmd;
	if (socket_server_host) delete[] socket_server_host;
	if (socket_server_port) delete[] socket_server_port;
	if (socket_auth_token) delete[] socket_auth_token;
	if (send_buf) delete[] send_buf;
	if (rcv_buf) delete[] rcv_buf;
	if (server_msg) delete[] server_msg;
	if (server_errno_msg) delete[] server_errno_msg;
	if (server_err_component) delete[] server_err_component;
	if (user_host) delete[] user_host;
	if (socket_file_name) delete[] socket_file_name;

}

// static initializers
const char* CoZBaseClient::SERVER_SUBSYS_NAME = "dspipes";

/*
 * Determine what protocol version to use when communicating with the server.
 * If the COZ_SERVER_PROTOCOL_VERSION environment variable is not set, use
 * COZ_PROTOCOL_VERSION.  Otherwise, if our version is greater than the server's
 * adopt the server's.  If the server's is greater, request that the server
 * use ours.
 * 
 * Release 4.5.0: Since we cannot determine the server protocol when initiating
 * remotely, the message log level has been lowered to DEBUG. 
 */
void CoZBaseClient::negotiateProtocolVersion() {
	char* p = getenv(ENV_COZ_SERVER_PROTOCOL_VERSION);
	if (!p) {
		LOG(LOG_DEBUG, "Server protocol version not available");
	} else {
		serverProtocolVersion = atoi(p);
		if (COZ_PROTOCOL_VERSION > serverProtocolVersion) {
			protocolVersion = serverProtocolVersion;
		}
	}
	LOG(LOG_DEBUG, "Protocol version server: %d, client: %d, negotiated: %d", serverProtocolVersion, COZ_PROTOCOL_VERSION, protocolVersion);
}

void CoZBaseClient::checkByteCounts() {
	if (!server_exitstatus_received) {
		LOG(LOG_DEBUG, "No exit status packet received, skipping byte count check.");
		return;
	}
	if (server_bytes_received != total_bytes_sent || server_bytes_sent != total_bytes_received) {
		LOG(LOG_ERR, "transmitted byte counts do not match, server sent/received = %llu/%llu, client received/sent = %llu/%llu",
			(unsigned long long)server_bytes_sent, (unsigned long long)server_bytes_received,
			(unsigned long long)total_bytes_received, (unsigned long long)total_bytes_sent);
		//If a return code has already been set, prefer that.  Otherwise set a system error.
		if (!return_code) {
			return_code = RC_SYSTEM_ERROR;
		}
	} else {
		LOG(LOG_DEBUG, "transmitted byte counts agree; server sent/received = %llu/%llu, client received/sent = %llu/%llu",
			(unsigned long long)server_bytes_sent, (unsigned long long)server_bytes_received,
			(unsigned long long)total_bytes_received, (unsigned long long)total_bytes_sent);
	}
}

bool CoZBaseClient::connectWithTimeout(struct sockaddr* address, socklen_t address_len) {
	TRACE_CTX("connectWithTimeout");

	int flags;
	fd_set rset, wset;
	struct timeval tv;
	flags = Fcntl(sock, F_GETFL, 0);
	Fcntl(sock, F_SETFL, flags | O_NONBLOCK);

	errno = 0;
	if (connect(sock, address, address_len) == 0) {
		Fcntl(sock, F_SETFL, flags);
		TRACE_CTX_RETVAL("true");
		return true;
	}
	if (errno != EINPROGRESS) {
		TRACE_CTX_RETVAL("false");
		return false;
	}
	FD_ZERO(&wset);
	FD_SET(sock, &wset);

	char* time_out = getenv(ENV_COZ_CONNECT_TIMEOUT_SEC);
	tv.tv_sec = time_out ? atoi(time_out) : DEFAULT_CONNECT_TIMEOUT_SEC;
	tv.tv_usec = 0;

	int rc = 0;
	if ((rc = Select(sock+1, NULL, &wset, NULL, &tv)) == 0) { // Wait for a writable socket
		LOG(LOG_WARNING, "Timeout attempting to connect socket");
		TRACE_CTX_RETVAL("false");
		return false;
	}
	TRACEV("select() returned: %d", rc);

	if (FD_ISSET(sock, &wset)) {
		struct sockaddr_storage paddr_in;
		socklen_t len = sizeof(struct sockaddr_storage);
		// determine if it really connected
		errno = 0;
		if (getpeername(sock, (struct sockaddr*)&paddr_in, &len) == -1) {
			TRACE_CTX_RETVAL("false");
			return false;
		}
	}
	Fcntl(sock, F_SETFL, flags);

	char paddr_out[MAX_PADDR_LEN];
	getPrintedSocketAddr(address, paddr_out);

	TRACE_CTX_RETVAL("true, %s", paddr_out);
	return true;
}


void CoZBaseClient::getPrintedSocketAddr(struct sockaddr* address, char* paddr) {
	struct sockaddr_in* a4 = (struct sockaddr_in*)address;
	struct sockaddr_in6* a6 = (struct sockaddr_in6*)address;
	switch (((struct sockaddr_in*)address)->sin_family) {
	case AF_INET:
		if (!inet_ntop(a4->sin_family, &a4->sin_addr, paddr, MAX_PADDR_LEN-1))
			strcpy(paddr, "?");
		break;
	case AF_INET6:
		if (!inet_ntop(a6->sin6_family, &a6->sin6_addr, paddr, MAX_PADDR_LEN-1))
			strcpy(paddr, "?");
		break;
	default:
		strcpy(paddr, "?");
	}
}

/**
 * exec()  - setup and exec remote command, handle IO, waitChild
 *
 */
int CoZClient::exec(const char* cmd_name, char** argv, int offset, int nargs) {

	const char* fn = "exec";
	TRACE_CTX_ARGS(fn, "offset=%d, nargs=%d", offset, nargs);

	pCoZClient = this;
	signal(SIGTERM, sig_term_handler);

	// we want to defer SIGPIPE signals and only (re-signal) then in specific cases
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
		LOG(LOG_ERR, "signal(SIGPPE, SIG_IGN) failed: %s", strerror(errno));
	}
	cmd_argc = nargs+1;
	cmd_argv = new const char*[cmd_argc+1];
	cmd_argv[0] = cmd_name;
	TRACEV("cmd_arg[0] = %s", cmd_argv[0]);
	cmd_argv[cmd_argc] = NULL;
	for (int i=1; i < cmd_argc; i++) {
		cmd_argv[i] = argv[i+offset-1];
		LOG(LOG_DEBUG,"cmd_arg[%d] = %s", i, cmd_argv[i]);
	}

	reserveStandardHandles();  // redirects STDIN/STDOUT/STDERR to /dev/null if their FDs are missing

    #ifdef TARGET_CYGWIN
	if (io_config_stdin == IO_CONFIG_TEXT) {
		setmode(STDIN_FD, O_TEXT);
	} else {
		setmode(STDIN_FD, O_BINARY);
	}

	if (io_config_stdout == IO_CONFIG_TEXT) {
		setmode(STDOUT_FD, O_TEXT);
	} else {
		setmode(STDOUT_FD, O_BINARY);
	}

	if (io_config_stderr == IO_CONFIG_TEXT) {
		setmode(STDERR_FD, O_TEXT);
	} else {
		setmode(STDERR_FD, O_BINARY);
	}
    #endif

	try {
		if (conn_type == CONN_TYPE_SSH) {
			startSshConnection();
		} else {
			getSocketEnvironment();
			startSocketConnection();
		}
		sendControlPackets();
		handleCmdIO();

		if (conn_type == CONN_TYPE_SSH) {
			return_code = waitChild();
		} else {
			// The return_code should have been received in a CMD_EXIT_CODE packet
			// received from the server.
			close(sock);
		}

		checkByteCounts();

	} catch (const int _rc) {
		return_code = _rc;
	}

	// don't override a return code if already set
	if (!server_exitcode_received && return_code == 0) {
		LOG(LOG_ERR, "Error: no exit code received from CoZServer");
		return_code = RC_SYSTEM_ERROR;
	}

	return return_code;
}

/**
 * Set the options for the ssh command, from args[iargs]
 * through args[iargs+nargs-1].  The last argument given
 * must be "user@host".
 *
 * This message is required to use CONN_TYPE_SSH
 */
void CoZClient::setSshOptions(char** args, int iargs, int nargs) {

	TRACE_CTX("setSshOptions");
	conn_type = CONN_TYPE_SSH;   // this means that we are using ssh!

	int argc = nargs + 100; //allow for additional ssh options to specified via env var.
	ssh_cmd = new char*[argc];

	// Get the command name and leading options
	if(char* p = getenv(ENV_COZ_SSH_CMD)) {
		char* command = new char[strlen(p)+1];
		strcpy(command,p);
		CoZUtil::parseCommand(command, ssh_cmd, argc);
	} else {
		argc = 0;
		ssh_cmd[argc++] = const_cast<char*>("ssh");
	}

	// Add the options given: args[istart] through args[istart+nargs-2]
	// (not user@host just yet)
	for (nargs+=iargs; iargs<nargs-1; iargs++) {
		ssh_cmd[argc++] = args[iargs];
	}

	// Get the name of the ssh subsystem to use
	char* subsys = getenv(ENV_COZ_SSH_SUBSYS);
	if (!subsys) {
		subsys = (char*)SERVER_SUBSYS_NAME;  // use the default
	}
	ssh_cmd[argc++] = const_cast<char*>("-s");

	ssh_cmd[argc++] = setUserHost(args[nargs-1]);  // user@host, adds user when @host specified

	ssh_cmd[argc++] = subsys;
	ssh_cmd[argc++] = NULL;  // last arg is NULL

	if (LOG_ACTIVE(LOG_TRACE)) {
		for (int i=0; ssh_cmd[i]; i++) {
			TRACEV("ssh_cmd[%d] = %s", i, ssh_cmd[i]);
		}
	}
}

/**
 * Copy input from stdin to the fd_conn_snd connection to CoZServer's stdin,
 * and copy and demux output packets from CoZServer's stdout (received on fd_conn_rcv)
 * to either stdout or stderr.
 *
 * So as not to allow one IO to block any others, a non-blocking select loop is used.
 *
 */
void CoZClient::handleCmdIO() {

	const char* fn = "handleCmdIO";
	TRACE_CTX_ARGS(fn, "fd_conn_snd=%d fd_conn_rcv=%d", fd_conn_snd, fd_conn_rcv);

	fd_set read_fds;
	fd_set write_fds;
	int maxfdp1 = ((fd_conn_snd > fd_conn_rcv) ? fd_conn_snd : fd_conn_rcv) + 1;

	// Configure all fds as nonblocking
	int flags = Fcntl(fd_conn_snd, F_GETFL, 0);
	Fcntl(fd_conn_snd, F_SETFL, flags | O_NONBLOCK);

	if (fd_conn_rcv != fd_conn_snd) { // different if pipes; same if socket
		flags = Fcntl(fd_conn_rcv, F_GETFL, 0);
		Fcntl(fd_conn_rcv, F_SETFL, flags | O_NONBLOCK);
	}

	bool stdin_drained = (io_config_stdin == IO_CONFIG_NONE);
	if (!stdin_drained && !isatty(STDIN_FD)) {
		flags = Fcntl(STDIN_FD, F_GETFL, 0);
		Fcntl(STDIN_FD, F_SETFL, flags | O_NONBLOCK);
		TRACE("set STDIN_FD to O_NONBLOCK");
	}

	bool stdout_drained = (io_config_stdout == IO_CONFIG_NONE);
	int stdout_flags = Fcntl(STDOUT_FD, F_GETFL, 0);   // may use these flags later
	if (!stdout_drained && !isatty(STDOUT_FD)) {
		Fcntl(STDOUT_FD, F_SETFL, stdout_flags | O_NONBLOCK);
		TRACE("set STDOUT_FD to O_NONBLOCK");
	}

	if (!send_buf) send_buf = new char[MAX_PACKET_LEN];
	if (!rcv_buf)  rcv_buf  = new char[MAX_PACKET_LEN];

	// we read from stdin and send to CoZServer on fd_conn_snd
	char* send_read_ptr = send_buf;
	char* send_write_ptr = send_buf;
	char* send_buf_end = send_buf + MAX_PACKET_LEN;

	// we received muxed stdout/stderr/status packets from CoZServer on fd_conn_rcv
	CoZPacket* packet = (CoZPacket*)rcv_buf;
	char* rcv_read_ptr =  (char*)packet;    // start reading the header
	char* rcv_write_ptr = packet->payload.raw;  // start writing the payload
	char* rcv_buf_end = rcv_buf + MAX_PACKET_LEN;

	ssize_t nRead = 0;
	ssize_t nWritten = 0;
	ssize_t bytesToWrite = 0;
	ssize_t bytesToRead = PACKET_HEADER_LEN;  // start by reading a header

	if (protocolVersion == 1 && stdin_drained) {
		// send EOF to server if we don't have stdin and we are running protocol 1
		if (shutdown(fd_conn_snd, SHUT_WR)) {
			LOG(LOG_WARNING, "shutdown(fd_conn_snd, SHUT_WR) error on no stdin: %s", strerror(errno));
		} else {
			TRACE("shutdown(fd_conn_snd, SHUT_WR) successful on no stdin");
		}
	}

	for (;;) {

		FD_ZERO(&read_fds);
		FD_ZERO(&write_fds);
		bool canReadStdin = false;  // used as extra guard against false FD_ISSET(STDIN_FD)
		if (!stdin_drained && send_read_ptr < send_buf_end) {
			TRACE("FD_SET: STDIN_FD");
			canReadStdin = true;
			FD_SET(STDIN_FD, &read_fds);
		}
		if (bytesToRead) {
			TRACEV("FD_SET: fd_conn_rcv, bytesToRead=%d", bytesToRead);
			FD_SET(fd_conn_rcv, &read_fds);
		}
		if (send_write_ptr < send_read_ptr) {
			TRACE("FD_SET: fd_conn_snd");
			FD_SET(fd_conn_snd, &write_fds);
		}
		if (!stdout_drained && rcv_write_ptr < rcv_read_ptr) {
			TRACE("FD_SET: STDOUT_FD");
			FD_SET(STDOUT_FD, &write_fds);
		}

		TRACE("select()...");
		int nfds;
		if ((nfds = select(maxfdp1, &read_fds, &write_fds, NULL, NULL)) == -1) {
			if (errno == EINTR && termination_requested) {
				//We have been sent a SIG_TERM.
				LOG(LOG_INFO, "%s: select() interrupted", fn);
				break;
			} else {
				LOG(LOG_ERR, "%s: select() error: %s", fn, strerror(errno));
				throw RC_IO_ERROR;
			}
		}
		TRACEV("select() returned %d", nfds);
		if (nfds == 0) continue;

		if (FD_ISSET(STDIN_FD, &read_fds)) {
			TRACE("select() STDIN_FD...");
			if (canReadStdin) {
				nRead = read(STDIN_FD, send_read_ptr, send_buf_end - send_read_ptr);
				if (nRead < 0) {
					if (errno != EWOULDBLOCK) {
						LOG(LOG_ERR, "%s: read error on stdin: %s", fn, strerror(errno));
						if (errno == EPIPE) {
							throwSigPipe();  // and rest of code would be unreachable
						}
						throw RC_IO_ERROR;
					}
				} else if (nRead == 0) {
					TRACE("EOF on stdin");
					stdin_drained = true;
					FD_SET(fd_conn_snd, &write_fds);  // encourage write attempt to fd_conn_snd
				} else {
					TRACEV("read %d bytes from stdin", nRead);
					total_bytes_sent += nRead;  // accum bytes sent to server
					send_read_ptr += nRead;
					FD_SET(fd_conn_snd, &write_fds);  // encourage write attempt to fd_conn_snd
				}
			}
		}

		if (FD_ISSET(fd_conn_rcv, &read_fds)) {
			TRACEV("select() fd_conn_rcv, bytesToRead=%d", bytesToRead);

			nRead = read(fd_conn_rcv, rcv_read_ptr, bytesToRead);
			if (nRead < 0) {
				if (errno != EWOULDBLOCK) {
					LOG(LOG_ERR, "%s: read error on fd_conn_rcv: %s", fn, strerror(errno));
					throw RC_IO_ERROR;
				}
			} else if (nRead == 0) {
				if (rcv_read_ptr > (char*)packet ) {
					LOG(LOG_ERR, "%s: error: EOF after receiving partial packet from CoZServer", fn);
					throw RC_IO_ERROR;
				} else {
					TRACE("EOF on fd_conn_rcv");
					bytesToRead = 0;  // don't try to read any more
					if (stdin_drained && (rcv_write_ptr >= rcv_read_ptr)) {
						break;  // we are done with both directions
					} else {
						LOG(LOG_ERR, "error: CoZServer terminated before all input was read");
						throw RC_IO_ERROR;
					}
				}
			} else {
				TRACEV("read %d bytes from fd_conn_rcv", nRead);
				rcv_read_ptr += nRead;
				bytesToRead -= nRead;
				if (rcv_read_ptr == packet->payload.raw) { // got the entire header
					packet->len = toNetUShort(packet->len);
					if (packet->len > MAX_PACKET_PAYLOAD ||
					  	(packet->type == PACKET_TYPE_DATA
					  		&& packet->code > STDERR_FD ) )
					{
					  	LOG(LOG_ERR, "error: invalid packet received from CoZServer; type=%d, len=%d, code=%d",
					  					packet->type, packet->len, packet->code);
					  	throw RC_IO_ERROR;
					} else {
						bytesToRead = packet->len;
						if (bytesToRead == 0) {
							// Handle the special case of a data packet with zero-length payload,
							// which we might eventually use.  Added after 1.2.1
							rcv_read_ptr =  (char*)packet;    // start reading the next header
							rcv_write_ptr = packet->payload.raw;  // start writing the next payload
							bytesToRead = PACKET_HEADER_LEN;
						}
						TRACEV("packet header received from CoZServer; type=%d, len=%d, code=%d",
					  					packet->type, packet->len, packet->code);
					}

				} else if (rcv_read_ptr > packet->payload.raw) { // got at least part of payload
					if (packet->type == PACKET_TYPE_CMD &&
						packet->len == (rcv_read_ptr - packet->payload.raw))
					{
						// got a complete command packet.
						handleReceivedCommandPacket(packet);
						rcv_read_ptr =  (char*)packet;    // start reading the next header
						rcv_write_ptr = packet->payload.raw;  // start writing at the next payload
						bytesToRead = PACKET_HEADER_LEN;
					} else {
						// got some data in a data packet
						FD_SET(STDOUT_FD, &write_fds);  // encourage write attempt below
					}
				}
			}
		}

		if ( (bytesToWrite = rcv_read_ptr - rcv_write_ptr) > 0
				&& (FD_ISSET(STDOUT_FD, &write_fds) || packet->code == STDERR_FD)
				&& packet->type == PACKET_TYPE_DATA )
		{
			TRACE("select() STDOUT_FD...");
			// We write the packet to either stdout or stderr, but
			// we have only selected & nonblocked stdout, so it could be
			// that stderr blocks, but large volumes send to stderr should
			// only be for debugging anyway, so we can live with it.
			nWritten = write(packet->code, rcv_write_ptr, bytesToWrite);
			if (nWritten < 0) {
				if (errno == EAGAIN || errno == EWOULDBLOCK) {
					// EAGAIN can happen if stdout is a pipe, even if selected in write_fds
					// Can't see how we would ever get EWOULDBLOCK (stdout would not be a socket)
					// We know that it must be stdout, since stderr is always blocking
					// We will flip stdout to blocking and redrive the write
					Fcntl(STDOUT_FD, F_SETFL, stdout_flags & ~O_NONBLOCK);
					nWritten = write(packet->code, rcv_write_ptr, bytesToWrite);
					if (nWritten < 0) {
						LOG(LOG_ERR, "%s: (blocking) write error on %s: %s", fn, (packet->code == 1 ?"stdout":"stderr"), strerror(errno));
						if (errno == EPIPE) {
							throwSigPipe();  // and rest of code would be unreachable
						}
						throw RC_IO_ERROR;
					}
					Fcntl(STDOUT_FD, F_SETFL, stdout_flags | O_NONBLOCK);
					goto write_stdfd_ok;
				} else {
					LOG(LOG_ERR, "%s: (nonblocking) write error on %s: %s", fn, (packet->code == 1 ?"stdout":"stderr"), strerror(errno));
					if (errno == EPIPE) {
						throwSigPipe();  // and rest of code would be unreachable
					}
					throw RC_IO_ERROR;
				}
			} else {
			write_stdfd_ok:
				if (packet->code == STDOUT_FD) {
					total_bytes_received += nWritten;  // accum total bytes rcved from server
					TRACEV("wrote %d bytes to stdout", nWritten);
				}
				rcv_write_ptr += nWritten;
				if (rcv_write_ptr == rcv_read_ptr &&
				    rcv_write_ptr == (packet->payload.raw + packet->len) )
				{
					// when a complete packet is both read and written,
					// then reset to read the next packet
					rcv_read_ptr =  (char*)packet;    // start reading the header
					rcv_write_ptr = packet->payload.raw;  // start writing the payload
					bytesToRead = PACKET_HEADER_LEN;

					if (packet->code == STDOUT_FD) {
						TRACEV("wrote complete %d byte packet to stdout", packet->len);
					}
				}
			}
		}

		if (FD_ISSET(fd_conn_snd, &write_fds)
				&& ((bytesToWrite = send_read_ptr - send_write_ptr) > 0 || stdin_drained))
		{
			TRACE("select() fd_conn_snd...");
			nWritten = write(fd_conn_snd, send_write_ptr, bytesToWrite);
			if (nWritten < 0) {
				if (errno != EWOULDBLOCK) {
					LOG(LOG_ERR, "%s: write error on fd_conn_snd: %s", fn, strerror(errno));
					throw RC_IO_ERROR;
				}
			} else {
				TRACEV("wrote %d bytes to fd_conn_snd", nWritten);
				send_write_ptr += nWritten;
				if (send_write_ptr == send_read_ptr) {
					send_write_ptr = send_read_ptr = send_buf;  // reset to beginning
					//If we were using stdin and it is finished tell the server
					if (stdin_drained) {
						if (conn_type == CONN_TYPE_SSH) {
							close(fd_conn_snd);
							TRACE("closed fd_conn_snd (stdin input to CoZServer)");
						} else if (protocolVersion == 1 || io_config_stdin != IO_CONFIG_NONE) { //Don't issue a shutdown if we didn't use stdin.
							if (shutdown(fd_conn_snd, SHUT_WR)) {
								LOG(LOG_WARNING, "shutdown(fd_conn_snd, SHUT_WR) error: %s", strerror(errno));
							} else {
								TRACE("shutdown(fd_conn_snd, SHUT_WR) successful");
							}
						}
					}
				}
			}
		}

	}  /* end of for (;;) select loop */

}  /* end of handleCmdIO() */


/**
 * Handle a command/status packet received from the remote server command
 */
void CoZClient::handleReceivedCommandPacket(CoZPacket* packet) {
	TRACE_CTX_ARGS("handleReceivedCommandPacket", "len=%d, code=%d",
					packet->len, packet->code);

	if (packet->code == CMD_CODE_EXIT_CODE) {
		server_exitcode_received = true;
		// only use the packet exit code if it is not zero, otherwise keep the
		// return code we already have, which might have otherwise been set.
		if (packet->payload.exit_code) {
			return_code = packet->payload.exit_code;
		}
		LOG(packet->payload.exit_code ? LOG_ERR: LOG_DEBUG,
				"server exit_code=%d", packet->payload.exit_code);

	} else if (packet->code == CMD_CODE_ERROR_STATUS) {
		// This packet encodes error / status information and
		// for some commands (todsn, fromdsn) is sent immediately
		// after the command opens the dataset (with a errno=0).
		server_exitstatus_received = true;
		server_retcode = toNetInt(packet->payload.error_status.retcode);
		server_errno  = toNetInt(packet->payload.error_status.errnum);
		server_errno2 = toNetInt(packet->payload.error_status.errno2);
		server_err_component = CoZUtil::newStringCopy(packet->payload.error_status.component);
		server_msg           = CoZUtil::newStringCopy(packet->payload.error_status.msg);
		server_errno_msg     = CoZUtil::newStringCopy(packet->payload.error_status.errno_msg);

		uint64_t srv_rcvd = toNetU64(packet->payload.error_status.bytes_received);
		uint64_t srv_sent = toNetU64(packet->payload.error_status.bytes_sent);
		if (srv_rcvd) server_bytes_received = srv_rcvd;
		if (srv_sent) server_bytes_sent = srv_sent;

		LOG(server_retcode ? LOG_ERR : LOG_DEBUG,
			"%s: %s, rc=%d, msg=\"%s\"",
			server_err_component,
			server_msg,
			server_retcode,
			server_errno_msg);

	} else if (packet->code == CMD_CODE_NOOP) {
		TRACE("received CMD_CODE_NOOP (keepalive)");
		
	} else {
		LOG(LOG_WARNING, "unknown command packet->code=%d, len=%d ignored",
						packet->code,
						packet->len);

	}


}


/**
 * Send the control and command packets to dtlcmd's stdin
 */
void CoZBaseClient::sendControlPackets() {

	TRACE_CTX("sendControlPackets");

	sendClientInfoPacket();
	sendIOConfigPacket();
	if (conn_type != CONN_TYPE_SSH) {
		sendAuthTokenPacket();
	}
	sendCommandPacket();
	sendStreamBeginsPacket();

}


/**
 * Send the "Client Info" control packet
 */
void CoZBaseClient::sendClientInfoPacket() {

	TRACE_CTX("sendClientInfoPacket");

	char pbuf[MAX_CONTROL_PACKET_LEN];
	memset(pbuf, 0, sizeof(pbuf));
	CoZPacket* packet = (CoZPacket*)pbuf;

	char  clientId[255];

	Setlocale(LC_ALL, "");

	//If the user has specified a custom client code page, prefer that to the current locale
	if(char* p = getenv(ENV_CLIENT_CODEPAGE)) {
		clientCodePage = new char[strlen(p)+1];
		strcpy(clientCodePage,p);
		LOG(LOG_DEBUG, "clientCodePage '%s' (via %s)", clientCodePage, ENV_CLIENT_CODEPAGE);
	} else {
		clientCodePage = nl_langinfo(CODESET);
		LOG(LOG_DEBUG, "clientCodePage '%s' (via setlocale)", clientCodePage);
	}

	errno = 0;

	iconv_t iconvCd = Iconv_open("UTF-8", clientCodePage);

	/* get the user@host id of the client */
	const char* logname = getlogin();
	if (!logname) {
		struct passwd *pw = getpwuid(getuid());
		if (pw) {
			logname = pw->pw_name;
		} else {
			logname = "user";
		}
	}
	Stringcopy(clientId, logname, sizeof(clientId));
	strncat(clientId, "@", sizeof(clientId)-strlen(clientId));
	size_t len = strlen(clientId);
	gethostname(clientId + len, sizeof(clientId)-len);
	len = strlen(clientId);

	packet->type = PACKET_TYPE_CMD;
	packet->code = CMD_CODE_CLIENT_INFO;

	packet->payload.client_info.version = toNetUShort(protocolVersion);

	#ifdef TARGET_CYGWIN
	packet->payload.client_info.client_opsys = CLIENT_OPSYS_WINDOWS;
	#else
	packet->payload.client_info.client_opsys = CLIENT_OPSYS_POSIX;
	#endif

	/* convert the clientId to null-terminated UTF-8 */
	char* srcBuf = clientId;
	char* targetBuf = packet->payload.client_info.client_id;
	size_t inBytesLeft = len;
	size_t outBytesLeft = 2 * inBytesLeft;
	Iconv(iconvCd, &srcBuf, &inBytesLeft, &targetBuf, &outBytesLeft);
	if (inBytesLeft) {
		throw RC_CONVERT_ERROR;
	}
	*targetBuf++ = '\0'; // null terminate and point to next byte

	/* convert the clientCodePage to null-terminated UTF-8 */
	srcBuf = clientCodePage;
	inBytesLeft = strlen(clientCodePage);
	outBytesLeft = 2 * inBytesLeft;
	Iconv(iconvCd, &srcBuf, &inBytesLeft, &targetBuf, &outBytesLeft);
	if (inBytesLeft) {
		LOG(LOG_ERR, "iconv(%s->UTF-8: %s) error:", strerror(errno));
		throw RC_CONVERT_ERROR;
	}
	*targetBuf++ = '\0'; // null terminate and point to next byte

	iconv_close(iconvCd);
	/* Store packet payload length (unsigned short big endian) */
	len  = targetBuf - (char*)(packet->payload.raw);

	packet->len = toNetUShort(len);

	ssize_t bytesWritten = Write(fd_conn_snd, (char*)packet, len+PACKET_HEADER_LEN);

	TRACEV("%d of %d bytes written", bytesWritten, len+PACKET_HEADER_LEN);

}


/**
 * Build and send the "IO Config" control packet
 */
void CoZBaseClient::sendIOConfigPacket() {

	TRACE_CTX("sendIOConfigPacket");

	char pbuf[MAX_CONTROL_PACKET_LEN];
	CoZPacket* packet = (CoZPacket*)pbuf;

	packet->type = PACKET_TYPE_CMD;
	packet->code = CMD_CODE_IO_CONFIG;

	packet->payload.io_config.std_in  = io_config_stdin;
	packet->payload.io_config.std_out = io_config_stdout;
	packet->payload.io_config.std_err = io_config_stderr;
	packet->payload.io_config.errpt  = io_config_errpt;

	packet->len = toNetUShort(sizeof(packet->payload.io_config));

	ssize_t bytesWritten = Write(fd_conn_snd,
								(char*)packet,
								sizeof(packet->payload.io_config)+PACKET_HEADER_LEN);
}

/**
 * Build and send the "Auth Token" control packet.
 * The auth token is sent in two cases:
 *   1) a Co:Z Launcher scenario when the
 *      COZ_AUTH_TOKEN environment variable
 *      contains a token.
 *   2) a cozcontrol scenario, when executing
 *      dspipes commands over a durable connection.
 */
void CoZBaseClient::sendAuthTokenPacket() {

	TRACE_CTX("sendAuthToken");

	char pbuf[MAX_CONTROL_PACKET_LEN];
	CoZPacket* packet = (CoZPacket*)pbuf;
	packet->type = PACKET_TYPE_CMD;
	packet->code = CMD_CODE_AUTH_TOKEN;
	packet->len = toNetUShort(AUTH_TOKEN_LEN);

	char* sToken = getenv(ENV_COZ_AUTH_TOKEN); // set for Co:Z Launcher
	if (!sToken) {
	    if (socket_auth_token) // set for a cozcontrol session
	        sToken = socket_auth_token;
	    else {
			LOG(LOG_DEBUG, "%s was not set", ENV_COZ_AUTH_TOKEN);
			return;
		}
	}

	if (strlen(sToken) != (2 * AUTH_TOKEN_LEN)) {
		LOG(LOG_WARNING, "%s=%s invalid; ignored", ENV_COZ_AUTH_TOKEN, sToken);
		return;
	}
	const char* hexChars = "0123456789ABCDEF";
	unsigned char byte = 0;
	for (int i=0; i<(2*AUTH_TOKEN_LEN); i++) {
		const char* p = strchr(hexChars, sToken[i]);
		if (!p) {
			LOG(LOG_WARNING, "%s=%s invalid; ignored", ENV_COZ_AUTH_TOKEN, sToken);
			return;
		}
		unsigned char nibble = p-hexChars;
		if (i%2) {
			byte = nibble;
		} else {
			byte = (byte<<4) | nibble;
			packet->payload.auth_token[i/2] = byte;
		}
	}

	ssize_t bytesWritten = Write(fd_conn_snd,
								(char*)packet,
								AUTH_TOKEN_LEN+PACKET_HEADER_LEN);

	return;

}

/**
 * Build and send the "Command" control packet
 *
 * Bug: this code will not allow a single quote to be passed to the target command:
 * For example: cozclient tso delete 'G1JAVA3.PRIVATE.RUNTST14.TODELETE' does not work.
 */
void CoZBaseClient::sendCommandPacket() {

	TRACE_CTX("sendCommandPacket");

	char pbuf[MAX_CONTROL_PACKET_LEN];
	CoZPacket* packet = (CoZPacket*)pbuf;

	packet->type = PACKET_TYPE_CMD;
	packet->code = CMD_CODE_COMMAND;

	// fill the packet->payload.command string with cmd_argv tokens.
	// the first arg (command name) is never quoted
	strcpy(packet->payload.command, cmd_argv[0]);
	strcat(packet->payload.command, " ");
	int len = strlen(cmd_argv[0]) + 1;  // accumulates length of command string

	// the remaining args are 'quoted' unless they contain a quote
	for (int i=1; i<cmd_argc; i++) {
		const char* q = strchr(cmd_argv[i],'\'');
		if (!q) {
			len += (strlen(cmd_argv[i]) + 3); /* add one for space character and 2 for single quotes */
		} else {
			len += (strlen(cmd_argv[i]) + 1); /* add one for space character */
		}
		if (len > sizeof(packet->payload.command)) {
			LOG(LOG_ERR, "remote command is too long");
			throw RC_INPUT_ERROR;
		}
		if (!q) strcat(packet->payload.command, "'");
		strcat(packet->payload.command, cmd_argv[i]);
		if (!q) strcat(packet->payload.command, "'");
		strcat(packet->payload.command, " ");
	}


	packet->len = toNetUShort(len);

	TRACEV("packet->payload.command=\"%s\" len=%d", packet->payload.command, len);

	ssize_t bytesWritten = Write(fd_conn_snd,
								(char*)packet,
								len+PACKET_HEADER_LEN);

}

/**
 * Build and send the "Stream begins" control packet
 */
void CoZBaseClient::sendStreamBeginsPacket() {

	TRACE_CTX("sendStreamBeginsPacket");

	char pbuf[MAX_CONTROL_PACKET_LEN];
	CoZPacket* packet = (CoZPacket*)pbuf;

	packet->type = PACKET_TYPE_CMD;
	packet->code = CMD_CODE_STREAM_BEGINS;

	packet->len = 0;

	ssize_t bytesWritten = Write(fd_conn_snd,
								(char*)packet,
								PACKET_HEADER_LEN);
}

/*
 * We have received a SIGTERM.  Force the socket (if we have one)
 * to be shutdown.
 */
void CoZClient::shutdownSocket() {
	TRACE_CTX("shutdownSocket");
	termination_requested = true;
	if (sock) {
		if (shutdown(sock, SHUT_RDWR)) {
			LOG(LOG_ERR, "shutdown() (%d) %s", errno, strerror(errno));
		}
		close(sock);
	}
}

/**
 * startSocketConnection()  - start a socket connection to CoZServer
 *
 * Sets fd_conn_snd and fd_conn_rcv to the input and output pipes to the
 * socket.
 */
void CoZClient::startSocketConnection() {

	TRACE_CTX("startSocketConnection");

#ifdef HAVE_GETADDRINFO
	// Get a list of addinfo's for a given hostname and service(port)
	LOG(LOG_DEBUG, "Using getaddrinfo() to start connection to server %s at port %s",
			socket_server_host, socket_server_port);
	struct addrinfo hints, *listAI, *nextAI;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	// if the socket_server_host is "NULL" or "null", then we instead use the
	// IPV4 and IPV6 loopback adapters returned by passing NULL to getaddrinfo().
	// This option is provided in order to be compatible with the way that SSHD binds and listens on loopback adapters,
	// but is not the recommended solution for most situations.   The recommended solution
	// is to move to OpenSSH 6.7 to get this: https://bugzilla.mindrot.org/show_bug.cgi?id=2222
	bool connect_loopback = !strcasecmp(socket_server_host, "NULL");

	Getaddrinfo(connect_loopback ? NULL : socket_server_host,
					socket_server_port, // or "port"
					&hints,
					&listAI);


	// Try to connect to each element in the list until success or exhausted
	nextAI = listAI;  // point to the first one
	do {
		errno = 0;
		sock = socket(nextAI->ai_family,
						nextAI->ai_socktype,
						nextAI->ai_protocol);
		if (sock < 0)
			continue;  // try the next one

		errno = 0;
		if (connectWithTimeout(nextAI->ai_addr, nextAI->ai_addrlen)) {
			break;  	// got one
		}
	} while ( (nextAI = nextAI->ai_next) != NULL);

	if (errno) {
		const char* reason = (errno == ENOTCONN) ? "timeout" : "failed";
		LOG(LOG_ERR, "%s %s: host=\"%s\", port=\"%s\": %s",
					TRACE_CTX_FN,
					reason,
					socket_server_host,
					socket_server_port,
					strerror(errno));
		close(sock);
		freeaddrinfo(listAI);
		throw RC_IO_ERROR;
	}

	freeaddrinfo(listAI);

#else
	LOG(LOG_DEBUG, "Using gethostbyname() to start connection to server %s at port %s",
			socket_server_host, socket_server_port);
	struct sockaddr_in servaddr;
	struct in_addr** pptr;
	struct in_addr*  inetaddrp[2];
	struct in_addr   inetaddr;
	struct hostent*  hp;
	if ((hp=gethostbyname(socket_server_host)) == NULL) {
		if (inet_aton(socket_server_host, &inetaddr) == 0) {
			LOG(LOG_ERR, "%s: IP address could not be determined for \"%s\"", socket_server_host);
			throw RC_INPUT_ERROR;
		} else {
			LOG(LOG_DEBUG, "Got ip address from inet_aton()");
			inetaddrp[0] = &inetaddr;
			inetaddrp[1] = NULL;
			pptr = inetaddrp;
		}
	} else {
		LOG(LOG_DEBUG, "Got ip address from gethostbyname()");
		pptr = (struct in_addr**) hp->h_addr_list;
	}

	for (; *pptr != NULL; pptr++) {
		sock = socket(AF_INET, SOCK_STREAM, 0);
		if (sock < 0)
			continue;  // try the next one

		bzero(&servaddr, sizeof(servaddr));
		servaddr.sin_family = AF_INET;
		servaddr.sin_port = htons(atoi(socket_server_port));
		memcpy(&servaddr.sin_addr, *pptr, sizeof(struct in_addr));

		if (sock < 0)
			continue;  // try the next one

		errno = 0;
		if (connectWithTimeout((struct sockaddr*)&servaddr, sizeof(sockaddr))) {
			break;  	// got one
		}
	}
	if (errno) {
		const char* reason = (errno == ENOTCONN) ? "timeout" : "failed";
		LOG(LOG_ERR, "%s %s: host=\"%s\", port=\"%s\": %s",
					TRACE_CTX_FN,
					reason,
					socket_server_host,
					socket_server_port,
					strerror(errno));
		close(sock);
		throw RC_IO_ERROR;
	}
#endif //HAVE_GETADDRINFO

	// Both send and receive fds are the socket fd
	fd_conn_snd = fd_conn_rcv = sock;

}

/**
 * startSshConnection()  - start an ssh connection.
 *
 * Sets fd_conn_snd and fd_conn_rcv to the input and output pipes to the
 * ssh command
 */
void CoZBaseClient::startSshConnection() {

	TRACE_CTX("startSshConnection");

	int pipe_stdin[2];  // pipe from us to child(ssh) stdin
	int pipe_stdout[2]; // pipe from child(ssh) stdout to us

	errno = 0;
	Pipe(pipe_stdin);
	Pipe(pipe_stdout);

    #ifdef TARGET_CYGWIN
	setmode (pipe_stdin[WRITE_END], O_BINARY);
	setmode (pipe_stdin[READ_END], O_BINARY);
	setmode (pipe_stdout[WRITE_END], O_BINARY);
	setmode (pipe_stdout[READ_END], O_BINARY);
    #endif

	// TODO do a local spawn if on z/os?
	errno = 0;
	ssh_pid = Fork();

	if (ssh_pid) {
		if (ssh_pid > 0) {
			/* close unused pipe sides in parent */
			errno = 0;
			Close(pipe_stdin[READ_END]);
			Close(pipe_stdout[WRITE_END]);

			fd_conn_snd = pipe_stdin[WRITE_END];
			fd_conn_rcv = pipe_stdout[READ_END];
		}
		return;  // parent is done
	}

	// Here only if in child/forked process

	/* close unused pipe sides in child */
	Close(pipe_stdin[WRITE_END]);
	Close(pipe_stdout[READ_END]);

	Dup2(pipe_stdin[READ_END], STDIN_FD);
	Dup2(pipe_stdout[WRITE_END], STDOUT_FD);

	execvp(ssh_cmd[0], ssh_cmd);
	/* should not get here */
	LOG(LOG_ERR, "execvp: %s", strerror(errno));
	exit(RC_SYSTEM_ERROR);

}


/**
 * Answer a network (big endian) unsigned short
 */
unsigned short CoZBaseClient::toNetUShort(unsigned short value)
{
   long int i = 1;
   char *p = (char *) &i;
   if (p[0] == 1) {
   	  // we are little endian, return bytes reversed
      return (value >> 8) | (value << 8);
   } else {
   	  // we are big endian, just return the value
      return value;
   }
}

/**
 * Answer a network (big endian) signed int
 * (or will take one and convert to a host int)
 */
int32_t CoZBaseClient::toNetInt(int32_t value)
{
   long int i = 1;
   char *p = (char *) &i;
   if (p[0] == 1) {
   	  // we are little endian, return bytes reversed
   	  char* p = (char*)&value;
   	  char tc = *p;
   	  *p = *(p+3);
   	  *(p+3) = tc;
   	  tc = *(p+1);
   	  *(p+1) = *(p+2);
   	  *(p+2) = tc;
      return value;
   } else {
   	  // we are big endian, just return the value
      return value;
   }
}

/**
 * Answer a network (big endian) signed int
 * (or will take one and convert to a host int)
 */
uint64_t CoZBaseClient::toNetU64(uint64_t value)
{
   long int i = 1;
   char *p = (char *) &i;
   if (p[0] == 1) {
   	  // we are little endian, return bytes reversed
   	  char* p = (char*)&value;

   	  char tc = *p;
   	  *p = *(p+7);
   	  *(p+7) = tc;

   	  tc = *(p+1);
   	  *(p+1) = *(p+6);
   	  *(p+6) = tc;

   	  tc = *(p+2);
   	  *(p+2) = *(p+5);
   	  *(p+5) = tc;

   	  tc = *(p+3);
   	  *(p+3) = *(p+4);
   	  *(p+4) = tc;

      return value;
   } else {
   	  // we are big endian, just return the value
      return value;
   }
}


int CoZBaseClient::waitChild() {

	TRACE_CTX("waitChild");
    int stat;
    int rc = RC_SYSTEM_ERROR;
    errno = 0;
    if (wait(&stat) == -1) {
    	LOG(LOG_ERR, "Error waiting for child process. %s", strerror(errno));
    } else if (WIFEXITED (stat)) {
    	rc = WEXITSTATUS (stat);
    } else if (WIFSIGNALED (stat)) {
    	LOG(LOG_ERR, "Child process received signal: %d", WTERMSIG (stat));
   	} else if (WIFSTOPPED (stat)) {
   		LOG(LOG_ERR, "Child process stopped on signal: %d", WSTOPSIG (stat));
	} else {
		LOG(LOG_ERR, "Unknown child wait status code = %i", stat);
    }
    LOG(LOG_TRACE,"rc=%d", rc);
    return rc;
}

/**
 * This function should terminate the current process with a SIGPIPE
 * and not return.   This is what we want to do for EPIPE errors on
 * standard handles, since we want to leave it up to the enclosing shell or
 * calling program to decide if it is an issue.  For bash and most
 * shells it is ignored.
 */
void CoZBaseClient::throwSigPipe() {
	TRACE_CTX("throwSigPipe");
	signal(SIGPIPE, SIG_DFL);  // restore default handler (terminate process)
	kill(getpid(), SIGPIPE);
	sleep(5);  // since signals are delivered asynchronously
}

/**
 * This function will reserve FDs 0,1,2 so that
 * in the case that they are closed (missing), we will point them to
 * "/dev/null" and avoid trouble in handling of I/O
 */
void CoZBaseClient::reserveStandardHandles() {
	TRACE_CTX("reserveStandardHandles");

	int nfd = open("/dev/null", O_RDWR);
	if (nfd < 0) {
		LOG(LOG_ERR, "Unable to open /dev/null: %s", strerror(errno));
		return;
	}
	struct stat statbuf;
	for (int fd=0; fd < 3; fd++) {
		if(nfd == fd) {
			TRACEV("Missing fd%d was redirected to /dev/null", fd);
			continue;
		}
		if (fstat(fd, &statbuf)) {
			if (errno == EBADF) {
				if (dup2(nfd, fd) < 0) {
					LOG(LOG_ERR, "dup2(%d, %d) error: (%d) %s", nfd, fd, errno, strerror(errno));
				} else {
					TRACEV("Missing fd%d was duped to /dev/null", fd);
				}
			} else {
				LOG(LOG_ERR, "fstat(fd%d) error: (%d) %s", fd, errno, strerror(errno));
			}
		}
	}
	if (nfd > 2) {
		close(nfd);
	}
}


/*
 * Save the string in a class var user_host
 * to be used as part of the local unix domain socket name
 * in cozcontrol scenarios. If @host is specified, add the
 * user name. Note that a z/OS user name can contain @ so
 * the @ host scenario can only be detected when there is
 * only one @ and it is the first character.
 */
char * CoZBaseClient::setUserHost(char* _user_host) {

	int len = strlen(_user_host);
	struct passwd *pw = NULL;

	if (!_user_host) return (user_host = NULL);

	if (_user_host &&
		 *_user_host == '@' &&
		 strlen(_user_host) > 1 &&
		 !strchr(_user_host+1, '@')) {
		pw = Getpwuid(geteuid());
		len = len + strlen(pw->pw_name);
	}

	user_host = new char[len+1];
	if (pw) {
		strcpy(user_host, pw->pw_name);
		strcat(user_host, _user_host);
	} else {
	    strcpy(user_host, _user_host);
	}
	return user_host;

}

/*
 *
 * Two commands can be sent to the cozcontrol daemon listening on
 * the local unix domain socket:
 *
 *    CMD_CODE_COZCONTROL_ENV_REQ - requests environment information
 *    about the dspipes server running on z/OS.
 *    This command is sent by local dspipes commands such as fromdsn,
 *    todsn, etc that are to be executed on z/OS.
 *
 *    CMD_CODE_COZCONTROL_STOP_REQ - notifies the cozcontrol daemon to shutdown.
 *    This command is only sent by the cozcontrol stop command. When the
 *    daemon receives this command, it shuts down the session with z/OS and
 *    exits.
 *
 *  Returns 0 if the command was successful. -1 if an error occurred.
 *
 */
int CoZBaseClient::sendCmdToLocalUnixDomainSocket(unsigned char code) {
	TRACE_CTX_ARGS("sendCmdToLocalUnixDomainSocket", "%d", code);

	int return_code = -1;
	int local_sock;

	char rcv_buf[MAX_PACKET_LEN];
	CoZPacket* packet = (CoZPacket*)rcv_buf;
	char* rcv_read_ptr =  (char*)packet;    // start reading the header
	char* rcv_buf_end = rcv_buf + MAX_PACKET_LEN;

	ssize_t nRead = 0;
	ssize_t bytesToRead = PACKET_HEADER_LEN;  // start by reading a header

	selectSocketFileName(false);

	struct sockaddr_un server_addr;
	socklen_t server_len = sizeof(server_addr);
	memset((char*) &server_addr, 0, sizeof(server_addr));
	server_addr.sun_family = AF_UNIX;

	strncpy(server_addr.sun_path, socket_file_name, sizeof(server_addr.sun_path));

	// Create a socket and bind it to the Unix datagram socket path
	if ((local_sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
		LOG(LOG_ERR, "socket() failed: %s", strerror(errno));
		return -1;
	}

	try {

		if (connect(local_sock, (sockaddr*)&server_addr, sizeof(server_addr))== -1) {
			LOG(LOG_ERR, "connect() failed for %s : %s", "server", strerror(errno));
			throw RC_IO_ERROR;
		}

		// Send command

		memset((char*) rcv_buf, 0, sizeof(rcv_buf));
		packet->type = PACKET_TYPE_CMD;
		packet->code = code;
		packet->len = 0; //request packets have no payload

		if (write(local_sock, rcv_buf, PACKET_HEADER_LEN) < 0) {
			LOG(LOG_ERR, "write() failed for %s : %s", "server", strerror(errno));
			throw RC_IO_ERROR;
		}

		if (code == CMD_CODE_COZCONTROL_STOP_REQ) return 0;
		// Otherwise, receive the command response

		memset((char*) rcv_buf, 0, sizeof(rcv_buf));

		for (;;) {

			nRead = read(local_sock, rcv_read_ptr, bytesToRead);

			if (nRead < 0) {
				LOG(LOG_ERR, "read() failed : %s", strerror(errno));
				throw RC_IO_ERROR;
			} else if (nRead == 0) {
				if (rcv_read_ptr > (char*)packet ) {
					LOG(LOG_ERR, "read() failed: EOF after receiving partial packet");
					throw RC_IO_ERROR;
				} else {
					bytesToRead = 0;  // don't try to read any more
					LOG(LOG_ERR, "read() failed: EOF, daemon terminated before all input was read");
					throw RC_IO_ERROR;
				}
			} else {
				TRACEV("read %d bytes ", nRead);
				rcv_read_ptr += nRead;
				bytesToRead -= nRead;
				if (rcv_read_ptr == packet->payload.raw) { // got the entire header
					TRACEV("packet header received from daemon; type=%d, code=%d, len=%d",
									packet->type, packet->code, packet->len);

					bytesToRead = packet->len;
				} else if (rcv_read_ptr > packet->payload.raw) { // got at least part of payload
					if (packet->type == PACKET_TYPE_CMD &&
						packet->len == (rcv_read_ptr - packet->payload.raw))
					{
						// got a complete command packet.

						if ((code == CMD_CODE_COZCONTROL_ENV_REQ) && packet->code == CMD_CODE_COZCONTROL_ENV) {

							TRACEV("complete command packet received from daemon; type=%d, len=%d, code=%d",
									packet->type, packet->len, packet->code);

							socket_server_host = CoZUtil::newStringCopy(packet->payload.env_variable.host);
							socket_server_port = CoZUtil::newStringCopy(packet->payload.env_variable.port);
							socket_auth_token  = CoZUtil::newStringCopy((char *)packet->payload.env_variable.auth_token);

							LOG(LOG_DEBUG, " ENV set %s:%s", socket_server_host, socket_server_port);

							break; //Done since the cozcontrol daemon only sends one env packet

						} else {
							LOG(LOG_ERR, "bad packet read from the daemon, %d for request code %d", packet->code, code);
							throw RC_IO_ERROR;
						}
					} else {
						LOG(LOG_ERR, "incomplete payload read %d", rcv_read_ptr - packet->payload.raw);
						throw RC_IO_ERROR;
					}
				}
			}
		} // end for loop

		return_code = RC_OK;

	} catch (const int _rc) {
		// capture errors so close on socket can be done below
		return_code = -1;
	}

	close(local_sock);

    return return_code;
}

/*
 *
 * The socket file is created by cozcontrol start. Commands,
 * including cozcontrol stop, use the following logic to
 * select a socket file for use:
 *   if COZ_CONTROL_SESSION is set, use it
 *   else search ~/.ssh for files matching pattern cozcontrol-*.sock
 *      if one match, use it,
 *      else write an error indicating that COZ_CONTROL_SESSION
 *           should be set.
 *
 * On successful completion, class var socket_file_name is set.
 */
char * CoZBaseClient::selectSocketFileName(bool quiet) {
    struct dirent *pDirent;
    DIR *pDir;
    char *p;
    int pattern_matches = 0;

	// select the file name set by the environment variable, if present
	p = getenv(ENV_COZ_CONTROL_SESSION);
	if (p) {
		LOG(LOG_DEBUG, "Environment variable (%s) value is %s", ENV_COZ_CONTROL_SESSION, p);
		setUserHost(p);
		if (user_host) {
			   setSocketFileName();
		} else {
			LOG(LOG_ERR, "Environment variable (%s) value is invalid ", ENV_COZ_CONTROL_SESSION);
			if (quiet) return NULL;
			throw RC_INPUT_ERROR;
		}
	} else 	{

		struct passwd *pw;
	    pw = Getpwuid(geteuid());
	    char * ssh_dir = new char[strlen(pw->pw_dir)+strlen("/.ssh")+1];
	    strcpy(ssh_dir, pw->pw_dir);
	    //Check whether the home directory is just a single slash (i.e. $HOME is root).
	    //In this case, append .ssh rather than /.ssh.
	    if (strlen(ssh_dir) == 1)
	    	strcat(ssh_dir, ".ssh");
	    else
	    	strcat(ssh_dir, "/.ssh");

	    // match pattern cozcontrol-*.sock in ~/.ssh
		pDir = opendir (ssh_dir);
		if (pDir == NULL) {
			LOG(LOG_ERR, "Cannot open directory %s", socket_file_name);
			if (quiet) return NULL;
			throw RC_INPUT_ERROR;
		}

		while ((pDirent = readdir(pDir)) != NULL) {
			p = strstr(pDirent->d_name,".sock");
			if (p) {
				p = strstr(pDirent->d_name,"cozcontrol-");
				if (p == pDirent->d_name) {
					if (pattern_matches == 0) {
						pattern_matches++;

						socket_file_name = new char[strlen(ssh_dir)+strlen(pDirent->d_name)+2];
						strcpy(socket_file_name, ssh_dir);
						strcat(socket_file_name, "/");
						strcat(socket_file_name, pDirent->d_name);

						//Set user_host from the selected socket file name
						char * s = strstr(pDirent->d_name,".sock");
						if (s) {
							int len = strlen(pDirent->d_name)-strlen("cozcontrol-")-strlen(".sock");
							user_host = new char[len + 1];
							strncpy(user_host, p+strlen("cozcontrol-"), len);
						} else {
							LOG(LOG_ERR, "unexpected error setting user_host from socket file name");
							if (quiet) return NULL;
							throw RC_INPUT_ERROR;
						}
					} else {
						pattern_matches++;
					}
				}
			}
		}
		closedir (pDir);

		if (pattern_matches == 0) {
			LOG(LOG_ERR, "cozcontrol session not started.  Use the cozcontrol start command to start a session.");
			if (quiet) return NULL;
			throw RC_INPUT_ERROR;
		} else if (pattern_matches > 1) {
			LOG(LOG_ERR, "Multiple cozcontrol sessions found. Set env variable COZ_CONTROL_SESSION=userid@host to select a session.");
			if (quiet) return NULL;
			throw RC_INPUT_ERROR;
		}
	}

	LOG(LOG_DEBUG, "Selected sock file %s ", socket_file_name);
	return socket_file_name;
}

/*
 * Set the socket file path to a name based on user@host
 * and the user's .ssh directory path.
 *
 * socket name pattern
 * cozcontrol-<user@host>.sock
 */
void CoZBaseClient::setSocketFileName() {

	char* ptr;
	struct passwd *pw;
	pw = Getpwuid(geteuid());

    char * buf = new char[MAX_SOCKET_PATH_NAME+1];
	int n = snprintf(buf, MAX_SOCKET_PATH_NAME, "%s%s.ssh/cozcontrol-%s.sock",
			pw->pw_dir, (strlen(pw->pw_dir)==1 && pw->pw_dir[0]=='/')?"":"/", user_host);

	if (n > 0 && n < MAX_SOCKET_PATH_NAME) {
		socket_file_name = new char[strlen(buf)+1];
		strcpy(socket_file_name, buf);
	} else {
		LOG(LOG_ERR, "Socket file name path too long. Error: %d ", n);
		socket_file_name = NULL;
	}
	delete[] buf;

}

/**
 * Constructor/ initializers
 */
CoZClient::CoZClient() :
		CoZBaseClient(),
		user_set_stderr_option(false)
{

}

/**
 * Destructor
 */
CoZClient::~CoZClient() {

}

void CoZClient::setUserSetStderrOption(bool b) {
	user_set_stderr_option = b;
}

/**
 * getSocketEnvironment()  - set socket_server_host and socket_server_port
 * based on the environment. If the COZ environment variables are NOT set,
 * check to see if the environment is available for client initiated dataset
 * pipes.
 */
void CoZClient::getSocketEnvironment() {

	TRACE_CTX("getSocketEnvironment");

	if (!socket_server_host) {
		char* p = getenv(ENV_COZ_SERVER_HOST);
		if (p) {
			socket_server_host = CoZUtil::newStringCopy(p);
		} else {
			if (sendCmdToLocalUnixDomainSocket(CMD_CODE_COZCONTROL_ENV_REQ) != 0) {
			   LOG(LOG_ERR, "%s: %s must be configured or a cozcontrol session must be started.", TRACE_CTX_FN, ENV_COZ_SERVER_HOST);
			   throw RC_INPUT_ERROR;
			}
		}
	}
	if (!socket_server_port) {
		char* p = getenv(ENV_COZ_SERVER_PORT);
		if (p) {
			socket_server_port = CoZUtil::newStringCopy(p);
		} else {
			LOG(LOG_ERR, "%s: %s not configured", TRACE_CTX_FN, ENV_COZ_SERVER_PORT);
			throw RC_INPUT_ERROR;
		}
	}

	//If client initiated dspipes and the user did not specify the -e option on the command,
	//set the -e default to text format (same as cozclient -et). This is
	//required because CoZClient-cmd.C will incorrectly have set the default to -es in this case.
	if (!user_set_stderr_option && socket_file_name) {
		setIOConfigStderr(IO_CONFIG_TEXT);
	}
}


// Remove :%p since this does not work with a tunnelled case when the port is not 22
const char* CoZControlClient::COZ_CONTROL_PATH = "~/.ssh/cm-%r@%h.sock";

/**
 * Constructor/ initializers
 */
CoZControlClient::CoZControlClient() :
		CoZBaseClient(),
		local_sock(0),
		foreground_pid(getpid()),
	    env_coz_server_host(NULL),
	    env_coz_server_port(NULL),
	    env_coz_auth_token(NULL),
		env_coz_control_path(NULL)
{

}

/**
 * Destructor
 */
CoZControlClient::~CoZControlClient() {
	TRACE_CTX_FINE("~CoZControlClient");
	if (env_coz_server_host) delete[] env_coz_server_host;
	if (env_coz_server_port) delete[] env_coz_server_port;
	if (env_coz_auth_token) delete[] env_coz_auth_token;
	if (env_coz_control_path) delete[] env_coz_control_path;

	// we are either a backgrounded cozcontrol start process
	// or stop command here, so we unlink the unix domain socket.
	// ignore ENOENT errors as the other side may have cleaned
	// up already.  The socket_file_name iv is deleted in
	// CoZBaseClient because it is used by standard CoZClient
	// commands to see what connection mode we are in
	if (socket_file_name) {
		if (unlink(socket_file_name) == -1) {
			if (errno != ENOENT) {
				LOG(LOG_ERR, "cozcontrol session socket file %s could not be removed", socket_file_name);
			}
		}
	}
}

/**
 * exec()  - handle the cozcontrol command which has the following functions:
 *              - start the dspipes subsystem on z/OS
 *              - stop the dspipes subsystem on z/OS
 *
 * start - the start command will first start a unix domain socket listener. This listener
 *         gates any dspipes commands attempting to use the durable connection.
 *         After starting the listener, an SSH connection is established to the target server,
 *         starting the dspipes subsystem which runs a socket server. Once the cozcontrol client
 *         exchanges information over the SSH connection (ensuring that the connection is working),
 *         the client exits, leaving a background process running the local unix domain socket
 *         listener. This background process runs until a cozcontrol stop command is issued or
 *         the SSH connection times out on the server side or is terminated. Note that the
 *         background process continues to run even if the user's terminal session ends.
 *         The background process responds to dspipes command requests for environment information
 *         about the cozcontrol session: hostname, port, auth token.
 *
 *         For the tunneled case, the client ssh connection is set up as a Master with a control path.
 *         When tunneling, the server side sets up a socket listener on 127.0.0.1:port where
 *         port is ephemeral. The client creates a local ephemeral port and sets up a local
 *         forward to the server:
 *            ssh -O forward -L localport:127.0.0.1:hostport -S ~/.ssh/cm-%r@%h.sock user@host
 *         Subsequent dspipes commands connect to 127.0.0.1:localport causing all traffic to go
 *         over the SSH connection.
 *
 *         Note: tunneling is not supported under CYGWIN on Windows because the control master
 *         options are not supported.
 *
 *
 * stop -  the stop command sends a message to the cozcontrol start unix domain socket server
 *         and exits. When cozcontrol start receives the stop message, it exits, terminating the
 *         ssh connection. When the ssh connection ends, the dspipes subsystem on z/OS terminates.
 *
 */
int CoZControlClient::exec(const char* cmd_name, char** argv, int offset, int nargs) {

	const char* fn = "exec_cozcontrol";
	TRACE_CTX_ARGS(fn, " offset=%d, nargs=%d", offset, nargs);

	pCoZClient = this;

	cmd_argc = nargs+1;
	cmd_argv = new const char*[cmd_argc+1];
	cmd_argv[0] = cmd_name;
	TRACEV("cmd_arg[0] = %s", cmd_argv[0]);
	cmd_argv[cmd_argc] = NULL;
	for (int i=1; i < cmd_argc; i++) {
		cmd_argv[i] = argv[i+offset-1];
		LOG(LOG_DEBUG,"cmd_arg[%d] = %s", i, cmd_argv[i]);
	}

	if (strcmp(cmd_name, "cozcontrol")) {
		LOG(LOG_ERR, "command cozcontrol expected");
		throw RC_SYSTEM_ERROR;
	}

	const char * cmd_arg = NULL;
	if (cmd_argc > 1) {
		cmd_arg = cmd_argv[1];
	} else {
		LOG(LOG_ERR, "a cozcontrol argument is required, one of start or stop.");
	    throw RC_SYSTEM_ERROR;
	}

	if (!strcmp(cmd_arg, "stop")) {
	
	    /* Send a stop message to the cozcontrol start unix domain socket listener
           and exit. When the backgrounded process receives the stop message, it will exit,
           terminating the ssh connection. When the ssh connection ends, the dspipes 
           subsystem on z/OS terminates.                                               */

		LOG(LOG_DEBUG, "Stopping a cozcontrol session...");
		if (sendCmdToLocalUnixDomainSocket(CMD_CODE_COZCONTROL_STOP_REQ) != 0) {
		   LOG(LOG_ERR, "failed to notify cozcontrol to stop for session %s", user_host);
		   throw RC_SYSTEM_ERROR;
		}
		return (RC_OK);

	}

	// process start command

	if (conn_type != CONN_TYPE_SSH) {
		LOG(LOG_ERR, "-ssh option is required for cozcontrol start");
	    throw RC_SYSTEM_ERROR;
	}

    #ifdef TARGET_CYGWIN
	if (io_config_stdin == IO_CONFIG_TEXT) {
		setmode(STDIN_FD, O_TEXT);
	} else {
		setmode(STDIN_FD, O_BINARY);
	}

	if (io_config_stdout == IO_CONFIG_TEXT) {
		setmode(STDOUT_FD, O_TEXT);
	} else {
		setmode(STDOUT_FD, O_BINARY);
	}

	if (io_config_stderr == IO_CONFIG_TEXT) {
		setmode(STDERR_FD, O_TEXT);
	} else {
		setmode(STDERR_FD, O_BINARY);
	}
    #endif

	try {

		local_sock = startUnixDomainSocketListener();

		/* Now that the local unix domain socket has been created,
		   subsequent dspipes commands will wait on the socket until
		   a select is done below in runUnixDomainSocketServer.
		 */

		forkToBackground(); // fork is issued and the parent exits when signaled by SIGUSR1 or SIGUSR2 below.

		signal(SIGTERM, sig_term_handler);
		signal(SIGINT,  sig_term_handler);

	    // Specify that the signal handler that takes three arguments
	    // instead of one, which is the default.
	    sVal.sa_flags = SA_SIGINFO;
	    sVal.sa_sigaction = sig_child_handler;
	    sigaction(SIGCHLD, &sVal, NULL);

		try {
			startSshConnection(); // starts the ssh connection as a child process
			sendControlPackets();
			handleCmdIO();
			
			if (tunneled) {
				char ssh_forward[1024];
				uint16_t ephemeral_port = 0;

				// attempt to forward over an ephemeral port. Try a maximum of three times. An attempt may fail if the port is in use.
				for (int attempts=1; attempts <= MAX_EPHEMERAL_PORT_ATTEMPTS; attempts++) {
					// create a socket with an ephemeral port
					ephemeral_port = getEphemeralPort();
					if (!ephemeral_port) {
						throw RC_SYSTEM_ERROR;
					}

					// setup local forwarding using the ephemeral port on the server (forcing interface IPV4 by specifying 127.0.0.1 as the first arg in the -L string)
					sprintf(ssh_forward, "ssh -O forward -L 127.0.0.1:%hu:127.0.0.1:%s -S %s %s", ephemeral_port, env_coz_server_port, env_coz_control_path, user_host);
					LOG(LOG_FINE, "try SSH local forward using ephemeral port %hu, attempts=%d", ephemeral_port, attempts);

					int result;
					result = system(ssh_forward);

					if (result == 0) {
						break;
					} else if (result == -1) {
						LOG(LOG_ERR,"system(%s) failed - [%d]%s", ssh_forward, errno, strerror(errno));
						throw RC_SYSTEM_ERROR;
					} else if (result > 0) {
						if (attempts == MAX_EPHEMERAL_PORT_ATTEMPTS) {
							LOG(LOG_ERR,"SSH forward failed: %s", ssh_forward);
							throw RC_SYSTEM_ERROR;
						} else {
							LOG(LOG_DEBUG,"SSH forward failed: %s", ssh_forward);
						}
					}
				}

		        LOG(LOG_DEBUG, "Local forward setup complete: %s", ssh_forward);
		    
		        // overwrite the server side ephemeral port so that dspipes commands will use the localhost:localport
		        sprintf(env_coz_server_port, "%hu", ephemeral_port);
            }
		} catch (const int _rc) {
			LOG(LOG_TRACE, "Failed to establish cozcontrol connection, signaling foreground process");
			if (kill(foreground_pid, SIGUSR2) == -1) {
				LOG(LOG_TRACE,"signal SIGUSR2 failed for foreground process (%d): [%d] %s", foreground_pid, errno, strerror(errno));
			}
			throw;
		}
		LOG(LOG_TRACE, "Established cozcontrol connection, signaling foreground process");
		if (kill(foreground_pid, SIGUSR1) == -1) {
			LOG(LOG_TRACE,"signal SIGUSR1 failed for foreground process (%d): [%d] %s", foreground_pid, errno, strerror(errno));
		}
        
		if (!termination_requested) {
			runUnixDomainSocketServer(local_sock);
		}

	} catch (const int _rc) {
		return_code = _rc;
	}

	if (!termination_requested) {
		termination_requested = true;
		kill_ssh_child();
	}

	TRACE_CTX_RETVAL("%d", return_code);
	return return_code;
}

/*
 * We have received a SIGTERM or SIGINT.  Remove the
 * unix domain socket file, if this process created it.
 */
void CoZControlClient::shutdownSocket() {
	TRACE_CTX("shutdownSocket");
	if (local_sock) {
		if (shutdown(local_sock, SHUT_RDWR)) {
			LOG(LOG_ERR, "shutdown() local sock (%d) %s", errno, strerror(errno));
		}
		close(local_sock);
	}
}

/**
 * Set the options for the ssh command, from args[iargs]
 * through args[iargs+nargs-1].  The last argument given
 * must be "user@host".
 *
 * This message is required to use CONN_TYPE_SSH
 * 
 */
void CoZControlClient::setSshOptions(char** args, int iargs, int nargs, bool tunneling) {

	TRACE_CTX("setSshOptions");
	conn_type = CONN_TYPE_SSH;   // this means that we are using ssh!

	int argc = nargs + 100; //allow for additional ssh options to specified via env var.
	ssh_cmd = new char*[argc];

	// Get the command name and leading options
	if(char* p = getenv(ENV_COZ_SSH_CMD)) {
		char* command = new char[strlen(p)+1];
		strcpy(command,p);
		CoZUtil::parseCommand(command, ssh_cmd, argc);
	} else {
		argc = 0;
		ssh_cmd[argc++] = const_cast<char*>("ssh");
	}

	// Add the options given: args[istart] through args[istart+nargs-2]
	// (not user@host just yet)
	for (nargs+=iargs; iargs<nargs-1; iargs++) {
		ssh_cmd[argc++] = args[iargs];
	}

	// Get the name of the ssh subsystem to use
	char* subsys = getenv(ENV_COZ_SSH_SUBSYS);
	if (!subsys) {
		subsys = (char*)SERVER_SUBSYS_NAME;  // use the default
	}

	if (tunneling) {
		tunneled = true;
		// if tunneling requested, set the -M and -S path options for the ssh command.
		ssh_cmd[argc++] = const_cast<char*>("-M");
		ssh_cmd[argc++] = const_cast<char*>("-S");
		// Get the control path name
		char* path = getenv(ENV_COZ_CONTROL_PATH);
		if (!path) {
			path = (char*)COZ_CONTROL_PATH;  // use the default
		}
		env_coz_control_path = new char[strlen(path)+1];
		strcpy(env_coz_control_path, path);
		ssh_cmd[argc++] = env_coz_control_path;

	}

	ssh_cmd[argc++] = const_cast<char*>("-s");

	ssh_cmd[argc++] = setUserHost(args[nargs-1]);  // user@host, adds user when @host specified

	ssh_cmd[argc++] = subsys;
	ssh_cmd[argc++] = NULL;  // last arg is NULL

	if (LOG_ACTIVE(LOG_TRACE)) {
		for (int i=0; ssh_cmd[i]; i++) {
			TRACEV("ssh_cmd[%d] = %s", i, ssh_cmd[i]);
		}
	}
}

/**
 * This is a simplified version of the CoZClient::handleCmdIO() method.
 *
 * On cozcontrol start, we receive one cmd packet from CoZServer on fd_conn_rcv.
 * Once this packet is received, the select ends. If the server never sends a
 * packet, the process will wait until the timeout occurs in the
 * z/OS dspipes subsystem. This timeout value is configurable in the
 * dspipes_config file z/OS.
 *
 */
void CoZControlClient::handleCmdIO() {

	const char* fn = "handleCmdIO";
	TRACE_CTX_ARGS(fn, "fd_conn_rcv=%d", fd_conn_rcv);

	fd_set read_fds;

	// Configure the rcv fd as nonblocking
	int flags = Fcntl(fd_conn_rcv, F_GETFL, 0);
	Fcntl(fd_conn_rcv, F_SETFL, flags | O_NONBLOCK);

	if (!rcv_buf)  rcv_buf  = new char[MAX_PACKET_LEN];

	CoZPacket* packet = (CoZPacket*)rcv_buf;
	char* rcv_read_ptr =  (char*)packet;    // start reading the header
	char* rcv_buf_end = rcv_buf + MAX_PACKET_LEN;

	ssize_t nRead = 0;
	ssize_t bytesToRead = PACKET_HEADER_LEN;  // start by reading a header

	for (;;) {

		FD_ZERO(&read_fds);

		if (bytesToRead) {
			TRACEV("FD_SET: fd_conn_rcv, bytesToRead=%d", bytesToRead);
			FD_SET(fd_conn_rcv, &read_fds);
		}

		TRACE("select()...");
		int nfds;
		if ((nfds = select(fd_conn_rcv+1, &read_fds, NULL, NULL, NULL)) == -1) {
			if (errno == EINTR && termination_requested) {
				//We have been sent a SIG_TERM.
				LOG(LOG_INFO, "%s: select() interrupted", fn);
				if (termination_requested)
					throw RC_IO_ERROR;
				break;
			} else {
				LOG(LOG_ERR, "%s: select() error: %s", fn, strerror(errno));
				throw RC_IO_ERROR;
			}
		}
		TRACEV("select() returned %d", nfds);
		if (nfds == 0) continue;

		if (FD_ISSET(fd_conn_rcv, &read_fds)) {
			TRACEV("select() fd_conn_rcv, bytesToRead=%d", bytesToRead);

			nRead = read(fd_conn_rcv, rcv_read_ptr, bytesToRead);
			if (nRead < 0) {
				if (errno != EWOULDBLOCK) {
					LOG(LOG_ERR, "%s: read error on fd_conn_rcv: %s", fn, strerror(errno));
					throw RC_IO_ERROR;
				}
			} else if (nRead == 0) {
				if (rcv_read_ptr > (char*)packet ) {
					LOG(LOG_ERR, "%s: error: EOF after receiving partial packet from CoZServer", fn);
					throw RC_IO_ERROR;
				} else {
					TRACE("EOF on fd_conn_rcv");
					bytesToRead = 0;  // don't try to read any more
					LOG(LOG_ERR, "error: CoZServer terminated before all input was read");
					throw RC_IO_ERROR;
				}
			} else {
				TRACEV("read %d bytes from fd_conn_rcv", nRead);
				rcv_read_ptr += nRead;
				bytesToRead -= nRead;
				if (rcv_read_ptr == packet->payload.raw) { // got the entire header
					packet->len = toNetUShort(packet->len);
					if (packet->len > MAX_PACKET_PAYLOAD ||
						(packet->type == PACKET_TYPE_DATA
							&& packet->code > STDERR_FD ) )
					{
						LOG(LOG_ERR, "error: invalid packet received from CoZServer; type=%d, len=%d, code=%d",
										packet->type, packet->len, packet->code);
						throw RC_IO_ERROR;
					} else {
						bytesToRead = packet->len;
						TRACEV("packet header received from CoZServer; type=%d, len=%d, code=%d",
										packet->type, packet->len, packet->code);
					}
				} else if (rcv_read_ptr > packet->payload.raw) { // got at least part of payload
					if (packet->type == PACKET_TYPE_CMD &&
						packet->len == (rcv_read_ptr - packet->payload.raw))
					{
						// got a complete command packet.
						handleReceivedCommandPacket(packet);

						rcv_read_ptr =  (char*)packet;    // start reading the next header
						bytesToRead = PACKET_HEADER_LEN;
						break;
					}
				}
			}
		}

	}  /* end of for (;;) select loop */

}

/**
 * Handle a command/status packet received from the remote server command
 */
void CoZControlClient::handleReceivedCommandPacket(CoZPacket* packet) {
	TRACE_CTX_ARGS("handleReceivedCommandPacket", "len=%d, code=%d",
					packet->len, packet->code);

	if (packet->code == CMD_CODE_COZCONTROL_ENV) {

		env_coz_server_host = CoZUtil::newStringCopy(packet->payload.env_variable.host);
		env_coz_server_port = CoZUtil::newStringCopy(packet->payload.env_variable.port);
		env_coz_auth_token  = CoZUtil::newStringCopy((char *)packet->payload.env_variable.auth_token);

		LOG(LOG_DEBUG, "%s=%s", packet->payload.env_variable.host, packet->payload.env_variable.port);

	} else if (packet->code == CMD_CODE_EXIT_CODE) {
		server_exitcode_received = true;
		LOG(packet->payload.exit_code ? LOG_ERR: LOG_DEBUG,
				"server exit_code=%d", packet->payload.exit_code);
		// only use the packet exit code if it is not zero, otherwise keep the
		// return code we already have, which might have otherwise been set.
		if (packet->payload.exit_code) {
			return_code = packet->payload.exit_code;
			throw return_code;
		}

	} else {
		LOG(LOG_ERROR, "unknown command packet->code=%d, len=%d system error",
						packet->code,
						packet->len);
		throw RC_SYSTEM_ERROR;
	}
}


/*
 * Run by the cozcontrol daemon to handle the following commands:
 *    CMD_CODE_COZCONTROL_ENV_REQ and CMD_CODE_COZCONTROL_STOP
 *
 * When CMD_CODE_COZCONTROL_ENV_REQ is received a response is written to
 * the client with socket connection and authtoken information.
 *
 * Note: Neither command has payload so the receive logic in this
 * method is simpler than other select/recv scenarios.
 *
 * The select will be interrupted when the SSH connection is terminated
 * either by a timeout on the server or a kill.
 *
*/

void CoZControlClient::runUnixDomainSocketServer(int listen_sock) {

   TRACE_CTX("runUnixDomainSocketServer");

   char buf[MAX_PACKET_LEN+1];

   fd_set readset;
   int result;
   int client_sock = -1;

   for (;;) {

	   FD_ZERO(&readset);
	   FD_SET(listen_sock, &readset);

	   result = select(listen_sock+1, &readset, NULL, NULL, NULL);

	   if (result < 0) {
		   if (errno == EINTR && termination_requested) {
				//We have been sent a SIG_TERM.
				LOG(LOG_ERR, "%s: connection ended - possibly due to server inactivity timeout or SSH connection termination", socket_file_name);
				throw RC_TARGET_KILLED;
		   }
		   LOG(LOG_ERR, "%s: select() failed : %s", socket_file_name, strerror(errno));
		   throw RC_IO_ERROR;
	   } else if (result == 0)  continue;
	   else if (result > 0 && FD_ISSET(listen_sock, &readset)) {

			if ((client_sock = accept(listen_sock, NULL, NULL)) < 0) {
				LOG(LOG_ERR, "%s: accept() failed : %s", socket_file_name, strerror(errno));
				throw RC_IO_ERROR;
			}

			int nRead = recv(client_sock, buf, PACKET_HEADER_LEN, 0);

			if (nRead < 0) {
				LOG(LOG_ERR, "%s: recv error : %s", socket_file_name, strerror(errno));
				throw RC_IO_ERROR;
			} else if (nRead == 0) {
				LOG(LOG_ERR, "error: EOF");
				throw RC_IO_ERROR;
			} else {
				if (nRead == PACKET_HEADER_LEN) {  // got the entire header (and packet since not expecting payload)

					CoZPacket* packet = (CoZPacket*)buf;

					if (packet->code == CMD_CODE_COZCONTROL_ENV_REQ) {

						// send ENV response
						memset((char*) buf, 0, sizeof(buf));
						packet->type = PACKET_TYPE_CMD;
						packet->code = CMD_CODE_COZCONTROL_ENV;
						packet->len = sizeof(packet->payload.env_variable);

						strcpy(packet->payload.env_variable.host, env_coz_server_host);
                        strcpy(packet->payload.env_variable.port, env_coz_server_port);
					    strcpy(packet->payload.env_variable.auth_token, env_coz_auth_token);

						size_t len = PACKET_HEADER_LEN + packet->len;

						if ((send(client_sock, buf, len, 0)) == -1) {
							LOG(LOG_ERR, "send() failed for %s: %s", socket_file_name, strerror(errno));
							throw RC_IO_ERROR;
						}

						LOG(LOG_DEBUG, "sent packet type/code/len %d/%d/%d  with %s:%s",
								packet->type, packet->code, packet->len, packet->payload.env_variable.host, packet->payload.env_variable.port);

					} else  {

						if (packet->code != CMD_CODE_COZCONTROL_STOP_REQ) {
							LOG(LOG_ERR, "Invalid request %d", packet->code);
							throw RC_SYSTEM_ERROR;
						}

						// exit to end the cozcontrol daemon
						LOG(LOG_INFO, "cozcontrol stop received. Exiting...");
						break;
					}
				}
			}
	   }
   }
}



/**
 * Started by the cozcontrol daemon to receive packets
 * from dspipes commands running in a client initiated
 * dspipes environment.
 */
int CoZControlClient::startUnixDomainSocketListener() {
	TRACE_CTX("startUnixDomainSocketListener");

	int listen_sock;

	setSocketFileName();

	struct sockaddr_un server_addr;
	socklen_t server_len = sizeof(server_addr);
	memset((char*) &server_addr, 0, sizeof(server_addr));
	server_addr.sun_family = AF_UNIX;

	strncpy(server_addr.sun_path, socket_file_name, sizeof(server_addr.sun_path));

	LOG(LOG_DEBUG, "using stream socket: \"%s\"", socket_file_name);

	// Create a socket and bind it to the socket path
	if ((listen_sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
		LOG(LOG_ERR, "socket() failed: %s", strerror(errno));
		throw RC_IO_ERROR;
	}

	if (bind(listen_sock, (sockaddr*)&server_addr, sizeof(server_addr))== -1) {
        if (EADDRINUSE==errno) {
        	LOG(LOG_ERR, "cozcontrol start already active (%s)", socket_file_name);
        } else {
        	LOG(LOG_ERR, "bind() failed for %s : %s (%d)", socket_file_name, strerror(errno), errno);
        }
		throw RC_IO_ERROR;
	}

	//chmod 600 on the socket filename
	if (chmod(socket_file_name, S_IRUSR|S_IWUSR) != 0) {
		LOG(LOG_ERR, "chmod() failed for %s : %s", socket_file_name, strerror(errno));
		throw RC_IO_ERROR;
	}

	if (listen(listen_sock, 10) < 0) {
		LOG(LOG_ERR, "listen() failed for %s : %s", socket_file_name, strerror(errno));
		throw RC_IO_ERROR;
	}

	return listen_sock;

}

/*
 * Allow the cozcontrol start program to run in the background. This is done
 * by issuing a fork and having the parent exit with a return code of 0.
 *
 * See the main comment for CoZControlClient::exec to understand why a background
 * process is needed.
 *
 * Note that once the cozcontrol start process is running in the background, it will not receive
 * a SIGHUP signal. If the terminal running the start command ends, the cozcontrol start
 * process and its child SSH process will continue to run until the timeout occurs in the
 * z/OS dspipes subsystem. This timeout value is configurable in the dspipes_config file
 * z/OS.
 *
 * setsid is not called by the child process, leaving the session id and process group id
 * unchanged for the child. The child process is not a true deamon process. It is a process
 * running in the background so setsid is not needed to ensure there is no controlling terminal.
 *
 */
void CoZControlClient::forkToBackground() {

	int 				i, fd0, fd1, fd2;
	pid_t 				pid;
	struct sigaction 	sa;

	/* capture the logging options */
	char* p_logging_options = getenv("COZ_LOG");

	/* Clear the file creation mask */
	umask(0);

	/* Issue a fork so that the same process is also running in the background. */
	if ((pid = fork()) < 0){
		LOG(LOG_ERR, "can't fork");
		throw RC_SYSTEM_ERROR;
	} else if (pid != 0) {
		// On the foreground side, setup and wait for a signal from the background side
		// indicating the success / failure of establishing the ssh connection.
		// We want to have a timeout here so that if the signal is not received in a
		// reasonable amount of time we can exit with a nice failure message.
		// It would be nice to be able to use sigtimedwait() for this, but some platforms
		// (cough, CYGWIN) don't support this function, so we use an alarm with sigwaitinfo instead.
		char* env_timeout = getenv(ENV_COZ_SSH_CONNECT_TIMEOUT_SEC);
		int timeout_secs = env_timeout ? atoi(env_timeout) : DEFAULT_COZ_SSH_CONNECT_TIMEOUT_SEC;

		sigset_t set;
		sigemptyset(&set);

		if(sigaddset(&set, SIGUSR1) == -1) {
			LOG(LOG_ERR, "Could not sigaddset() SIGUSR1");
			throw RC_SYSTEM_ERROR;
		}
		if(sigaddset(&set, SIGUSR2) == -1) {
			LOG(LOG_ERR, "Could not sigaddset() SIGUSR2");
			throw RC_SYSTEM_ERROR;
		}
		if(sigaddset(&set, SIGALRM) == -1) {
			LOG(LOG_ERR, "Could not sigaddset() SIGALRM");
			throw RC_SYSTEM_ERROR;
		}
		sigprocmask(SIG_BLOCK, &set, NULL);
		int signum;
		errno = 0;
		alarm(timeout_secs);
		if((signum = sigwaitinfo(&set, NULL)) == -1) {
			LOG(LOG_ERR, "sigwaitinfo() failed: [%d] %s", errno, strerror(errno));
			throw RC_SYSTEM_ERROR;
		} else if (signum == SIGUSR1) {
			LOG(LOG_DEBUG, "Established cozcontrol connection");
			exit(0);
		} else if (signum == SIGUSR2) {
			LOG(LOG_ERROR, "Failed to establish cozcontrol connection");
			exit(1);
		} else if (signum == SIGALRM) {
			LOG(LOG_ERR, "Timeout occurred trying to establish cozcontrol connection");
			kill(pid, SIGTERM); // Make sure the background process is killed
			exit(2);
		}
	}
	
	/* set the logging options for the deamon process */
	if (p_logging_options) {
		DtlLog::handleLogOption(p_logging_options);
	}

	/*
	 * Change the current working directory to the root
	 * so we won't prevent file systems from being unmounted.
	 */
	if (chdir("/") < 0) {
		LOG(LOG_ERR, "can't change dir to /");
		throw RC_SYSTEM_ERROR;
	}

}

/**
 * Get an ephemeral port for interface 127.0.0.1.
 * Returns ephemeral port, returns 0 on error.
 */
uint16_t CoZControlClient::getEphemeralPort() {
	TRACE_CTX("getEphemeralPort");

	const char* socket_server_host="127.0.0.1";
	const char* socket_server_port="0";
	uint16_t ephemeral_port = 0;
	int listenSock = 0;  // fd of socket - will be closed before return.

#ifdef HAVE_GETADDRINFO

	LOG(LOG_DEBUG, "Using getaddrinfo() to create socket to server %s at port %s",
			socket_server_host, socket_server_port);
	// Get a list of addinfo's for a given hostname and service(port)
	struct addrinfo hints, *listAI, *nextAI;
	memset(&hints, 0, sizeof(hints));
	hints.ai_flags  = AI_PASSIVE;
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	errno = 0;
	Getaddrinfo(socket_server_host,
					socket_server_port,
					&hints,
					&listAI);

	// Try to connect to each element in the list until success or exhausted
	nextAI = listAI;  // point to the first one
	do {
		errno = 0;
		listenSock = socket(nextAI->ai_family,
						nextAI->ai_socktype,
						nextAI->ai_protocol);
		if (listenSock < 0)
			continue;  // try the next one

		errno = 0;
		int opt = 1;  // "on"
		Setsockopt(listenSock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));

		errno = 0;
		if (bind(listenSock, nextAI->ai_addr, nextAI->ai_addrlen)) {
			close(listenSock); // error connecting, close and continue
		} else {
			break;  	// got one
		}
	} while ( (nextAI = nextAI->ai_next) != NULL);

	if (!nextAI) {
		LOG(LOG_ERR, "%s failed: %s, %s: %s",
					TRACE_CTX_FN,
					socket_server_host,
					socket_server_port,
					strerror(errno));
		freeaddrinfo(listAI);
		return 0;
	}


	char paddr[256];
	struct sockaddr_in * paddr_in = (struct sockaddr_in*)(nextAI->ai_addr);

	// Get the ephemeral port from the socket
	struct sockaddr_in sa;
	socklen_t len = sizeof(sa);
	if (getsockname(listenSock, (struct sockaddr *)&sa, &len) != 0) {
		LOG(LOG_ERR, "Unable to determine assigned ephemeral port");
		return 0;
	} else {
		ephemeral_port = ntohs(sa.sin_port);
	}
	LOG(LOG_DEBUG, "socket bound to ephemeral port: %s:%hu",
					inet_ntop(paddr_in->sin_family, &paddr_in->sin_addr, paddr, sizeof(paddr)-1),
					ephemeral_port);

	freeaddrinfo(listAI);

#else
	LOG(LOG_DEBUG, "Using gethostbyname() to create socket to server %s at port %s",
			socket_server_host, socket_server_port);
	struct sockaddr_in servaddr;
	struct in_addr** pptr;
	struct in_addr*  inetaddrp[2];
	struct in_addr   inetaddr;
	struct hostent*  hp;
	if ((hp=gethostbyname(socket_server_host)) == NULL) {
		if (inet_aton(socket_server_host, &inetaddr) == 0) {
			LOG(LOG_ERR, "%s: IP address could not be determined for \"%s\"", socket_server_host);
			throw RC_INPUT_ERROR;
		} else {
			LOG(LOG_DEBUG, "Got ip address from inet_aton()");
			inetaddrp[0] = &inetaddr;
			inetaddrp[1] = NULL;
			pptr = inetaddrp;
		}
	} else {
		LOG(LOG_DEBUG, "Got ip address from gethostbyname()");
		pptr = (struct in_addr**) hp->h_addr_list;
	}

	for (; *pptr != NULL; pptr++) {
		listenSock = socket(AF_INET, SOCK_STREAM, 0);
		if (listenSock < 0)
			continue;  // try the next one

		bzero(&servaddr, sizeof(servaddr));
		servaddr.sin_family = AF_INET;
		servaddr.sin_port = htons(atoi(socket_server_port));
		memcpy(&servaddr.sin_addr, *pptr, sizeof(struct in_addr));

		if (listenSock < 0)
			continue;  // try the next one

		errno = 0;
		int opt = 1;  // "on"
		Setsockopt(listenSock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));

		errno = 0;
		if (bind(listenSock, (sockaddr*)&servaddr, sizeof(servaddr))) {
			close(listenSock); // error connecting, close and continue
		} else {
			break;  	// got one
		}
	}
	if (errno) {
		LOG(LOG_ERR, "%s: host=\"%s\", port=\"%s\": %s",
					TRACE_CTX_FN,
					socket_server_host,
					socket_server_port,
					strerror(errno));
		close(sock);
		return 0;
	}

	// Get the ephemeral port from the socket
	struct sockaddr_in sa;
	socklen_t len = sizeof(sa);
	if (getsockname(listenSock, (struct sockaddr *)&sa, &len) != 0) {
		LOG(LOG_ERR, "Unable to determine assigned ephemeral port");
		return 0;
	} else {
		ephemeral_port = ntohs(sa.sin_port);
	}
	LOG(LOG_DEBUG, "socket bound to ephemeral port: %hu",
					ephemeral_port);


#endif //HAVE_GETADDRINFO


	// close the socket now that the ephemeral port has been selected (because only one bind is permitted on the port)
	close(listenSock);

	TRACE_CTX_RETVAL("ephemeral_port=%hu", ephemeral_port);
	return ephemeral_port;
}

